from .GraphEntry import GraphEntry
from ..enum_defines import TensorType


class ConstantEntry(GraphEntry):
    def __init__(self, name, tensor):
        super().__init__(name, tensor=tensor, is_constant_tensor=True)


class ScalarEntry(GraphEntry):
    def __init__(self, name, tensor):
        super().__init__(name, tensor=tensor, is_constant_tensor=True)
        self.tensor.shape = ()


class InputEntry(GraphEntry):
    def __init__(self, name, tensor):
        super().__init__(
            name, tensor=tensor, is_constant_tensor=False, tensor_type=TensorType.Input
        )


class OutputEntry(GraphEntry):
    def __init__(self, name, tensor):
        super().__init__(
            name, tensor=tensor, is_constant_tensor=False, tensor_type=TensorType.Output
        )


class EmptyEntry(GraphEntry):
    def __init__(self):
        super().__init__(name="null" + str(id(self)))
        self.set2_emptry()


__all__ = ["ConstantEntry", "ScalarEntry", "InputEntry", "OutputEntry", "EmptyEntry"]



import numpy as np
from typing import List

from ....graph import Tensor
from ....enum_defines import Layout, ALIGN16, ALIGN8, ALIGN2


class LayoutPerfData:
    def __init__(self, kernel: str = None, performance: int = 10):
        self.kernel = kernel
        self.performance = performance
        self.inputs_layout = []
        self.outputs_layout = []


class BaseLayout(object):
    def get_layout_perf_venus(self) -> List[LayoutPerfData]:
        perfs = []
        perf = LayoutPerfData()
        for tensor in self.inputs:
            perf.inputs_layout.append(Layout.NCHW)
        for tensor in self.outputs:
            perf.outputs_layout.append(Layout.NCHW)
        perfs.append(perf)
        return perfs

    def get_layout_perf(self) -> List[LayoutPerfData]:
        perfs = self.get_layout_perf_venus()
        return perfs


class CPUConvLayout(BaseLayout):
    def get_layout_perf_venus(self) -> List[LayoutPerfData]:
        perfs = []
        input = self.inputs[0]
        weight = self.inputs[1]

        kernel_num = weight.shape[0]
        kernel_c = weight.shape[1]

        group = self.attrs["group"]
        strides = self.attrs["strides"]
        stride_h = strides[0]
        if len(strides) == 1:
            stride_w = 1
        else:
            stride_w = strides[1]

        kernels = self.attrs["kernel_shape"]
        kernel_h = kernels[0]
        if len(kernels) == 1:
            kernel_w = 1
        else:
            kernel_w = kernels[1]

        h = input.shape[2]
        if len(input.shape) == 3:
            w = 1
        else:
            w = input.shape[3]

        data_size = (
            ALIGN8(kernel_c)
            * ((w + 8 * stride_w - 1) // (8 * stride_w))
            * (8 * stride_w)
            * h
        )
        data_size1 = (
            ALIGN8(kernel_c)
            * ((w + 8 * stride_w - 1) // (8 * stride_w))
            * (8 * stride_w)
        )
        data_size2 = ALIGN8(kernel_c) * h

        if group == kernel_c and group == kernel_num:  # depthwise conv
            num_input_align = ALIGN16(kernel_c)
            kernel_size = num_input_align * kernel_h * kernel_w
            assert (
                kernel_size * weight.dtype.itemsize <= 32768
            ), "kernel size of depthwise conv must less 32KB"
        elif group != 1:  # group conv
            assert kernel_num % group == 0
            num_input_align = ALIGN8(kernel_c)
            num_output_align = ((kernel_num // group + 1) // 2) * 2
            kernel_size = num_input_align * num_output_align * kernel_h * kernel_w
        else:  # common conv
            num_input_align = ALIGN8(kernel_c)
            num_output_align = ALIGN2(kernel_num)
            kernel_size = num_input_align * num_output_align * kernel_h * kernel_w

        perf = LayoutPerfData()
        if data_size > 65536 and kernel_size <= 32768:
            if data_size1 > 65536 and data_size2 <= 65536:
                for i in self.inputs:
                    perf.inputs_layout.append(Layout.NCWH)
                for i in self.outputs:
                    perf.outputs_layout.append(Layout.NCWH)
                perf.performance = 5
            elif data_size1 > 65536 and data_size2 > 65536:
                raise ValueError(f"data size exceed limit {data_size}")
            else:
                for i in self.inputs:
                    perf.inputs_layout.append(Layout.NCHW)
                for i in self.outputs:
                    perf.outputs_layout.append(Layout.NCHW)
        else:
            for i in self.inputs:
                perf.inputs_layout.append(Layout.NCHW)
            for i in self.outputs:
                perf.outputs_layout.append(Layout.NCHW)
        perfs.append(perf)

        return perfs


class CPUPoolLayout(BaseLayout):
    def get_layout_perf_venus(self) -> List[LayoutPerfData]:
        perfs = []
        input = self.inputs[0]

        kernel_num = input.shape[0]
        kernel_c = input.shape[1]

        strides = self.attrs["strides"]
        stride_h = strides[0]
        if len(strides) == 1:
            stride_w = 1
        else:
            stride_w = strides[1]

        h = input.shape[2]
        if len(input.shape) == 3:
            w = 1
        else:
            w = input.shape[3]

        data_size = (
            ALIGN8(kernel_c)
            * ((w + 8 * stride_w - 1) // (8 * stride_w))
            * (8 * stride_w)
            * h
        )
        data_size1 = (
            ALIGN8(kernel_c)
            * ((w + 8 * stride_w - 1) // (8 * stride_w))
            * (8 * stride_w)
        )
        data_size2 = ALIGN8(kernel_c) * h

        perf = LayoutPerfData()
        if data_size > 65536:
            if data_size1 > 65536 and data_size2 <= 65536:
                for i in self.inputs:
                    perf.inputs_layout.append(Layout.NCWH)
                for i in self.outputs:
                    perf.outputs_layout.append(Layout.NCWH)
                perf.performance = 5
            elif data_size1 > 65536 and data_size2 > 65536:
                raise ValueError(f"data size exceed limit {data_size}")
            else:
                for i in self.inputs:
                    perf.inputs_layout.append(Layout.NCHW)
                for i in self.outputs:
                    perf.outputs_layout.append(Layout.NCHW)
        else:
            for i in self.inputs:
                perf.inputs_layout.append(Layout.NCHW)
            for i in self.outputs:
                perf.outputs_layout.append(Layout.NCHW)
        perfs.append(perf)

        return perfs


__all__ = [
    "LayoutPerfData",
    "BaseLayout",
    "CPUConvLayout",
    "CPUPoolLayout",
]



import math
import numpy as np

from ...graph import Tensor
from ...enum_defines import DevType
from .base import Operator, register_op


@register_op
class iqVar(Operator):
    def infer_tensor(self):
        inputs = self.inputs
        X = inputs[0]
        x_shape = list(X.shape)
        x_shape[2] = 1

        scale_x = self.attrs.get("scale_x")
        temp = math.log(scale_x, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert self.inputs[0].scale == temp

        scale_o = self.attrs.get("scale_o")
        temp = math.log(scale_o, 2)
        assert abs(temp - int(temp)) < 0.000001
        self.outputs[0].scale = temp

        Y = X.clone(shape=tuple(x_shape), dtype=X.dtype, scale=int(temp))
        self.outputs[0] = Y

    def get_workspace(self, dev_type: DevType):
        workspace_size = 0
        workspace_size += (
            self.inputs[0].nbytes
            * self.inputs[0].dtype.itemsize
            / self.inputs[0].shape[2]
            * 3
        )
        max_workspace = Tensor.from_shape([workspace_size], np.int32, dev_type)
        return [max_workspace]


__all__ = ["iqVar"]



# Copyright (C) 2022 listenai Co.Ltd
# All rights reserved. 
# Created by leifang on 2022.09.31

import numpy as np
from typing import List, Dict, Tuple

from .graph import Graph, GraphEntry
from .graph_optimizer import op_split
from .enum_defines import MemType, DevType, Platform
from .resource_packer.memory import memory_plan, get_memory_size


def _get_next_id(threshold: List[int]) -> int:
    begin = len(threshold)
    end = 0
    for i in range(len(threshold)):
        if threshold[i] == 1:
            begin = i
        elif threshold[i] == -1:
            end = i
            break
    assert begin < end
    return (end + begin) // 2


def _judge_end(threshold: List[int]) -> int:
    for i in range(len(threshold) - 2):
        if threshold[i] == 1 and threshold[i + 1] == -1:
            return i
    return None


def _get_memory_plan(
    graph: Graph, platform: Platform, local_mem: str, threshold: int, is_dump: bool
) -> Tuple[Graph, Dict[int, List[int]], bool]:
    new_graph, is_linearint = op_split(graph, is_dump, threshold=threshold)
    new_graph.init_tensor()
    new_graph.update()
    new_graph = _graph_bind_device(new_graph, platform, local_mem)
    new_graph.pack_params()
    memory_planer = memory_plan(new_graph, True)
    return new_graph, memory_planer, is_linearint


def _graph_bind_device(graph: Graph, platform: Platform, memory: str) -> Graph:
    cpus = Platform.get_cpu_list(platform)
    if len(cpus) > 1:
        cpus = cpus[::-1]

    for node in graph.nodes.values():
        flag = 0
        if node.dev_type == None:
            for i in range(len(cpus)):
                if node.op_type in Platform.get_support_op(platform):
                    node.dev_type = cpus[i]
                    flag = 1
                    break
            if flag == 0:
                raise ImportError(f"platform:{platform} do not support {node.op_type}!")

        for i in range(len(node.inputs)):
            if node.inputs[i].is_constant():
                node.inputs[i].tensor.mem_type = MemType.from_str(memory)
            elif node.inputs[i] in graph.inputs:
                node.inputs[i].tensor.mem_type = MemType.SHARE_MEM
            elif node.inputs[i].tensor.mem_type == None:
                node.inputs[i].tensor.mem_type = MemType.SHARE_MEM

        for i in range(len(node.outputs)):
            if node.outputs[i].tensor.mem_type == None:
                node.outputs[i].tensor.mem_type = MemType.SHARE_MEM

    return graph


def graph_adapter(
    graph: Graph, platform: Platform, local_mem: str, is_dump: bool
) -> Tuple[Graph, Dict[int, List[int]]]:
    MIN_THRESHOLD = 32 * 1024
    MAX_THRESHOLD = 640 * 1024
    STEP_THRESHOLD = 10 * 1024
    threshold = np.arange(MIN_THRESHOLD, MAX_THRESHOLD, STEP_THRESHOLD)
    threshold_mask = np.zeros(len(threshold))
    print("set linearint threshold and analyze memory begin")
    print("try threshold:{}".format(threshold[-1]))

    new_graph, memory_planer, is_linearint = _get_memory_plan(
        graph, platform, local_mem, threshold[-1], is_dump
    )
    memory_tobe_allocated = get_memory_size(memory_planer, MemType.SHARE_MEM)

    if memory_tobe_allocated > MAX_THRESHOLD:
        threshold_mask[-1] = -1
        print("try threshold:{}".format(threshold[0]))
        new_graph, memory_planer, is_linearint = _get_memory_plan(
            graph, platform, local_mem, threshold[0], is_dump
        )
        memory_tobe_allocated = get_memory_size(memory_planer, MemType.SHARE_MEM)

        if memory_tobe_allocated > MAX_THRESHOLD:
            print(
                "WARNING:SHARE-MEM to be allocated was {}, exceed 640KB".format(
                    memory_tobe_allocated
                )
            )
        else:
            threshold_mask[0] = 1
            while is_linearint:
                id = _get_next_id(threshold_mask)
                print("try threshold:{}".format(threshold[id]))
                new_graph, memory_planer, is_linearint = _get_memory_plan(
                    graph, platform, local_mem, threshold[id], is_dump
                )
                memory_tobe_allocated = get_memory_size(
                    memory_planer, MemType.SHARE_MEM
                )

                if memory_tobe_allocated > MAX_THRESHOLD:
                    threshold_mask[id] = -1
                else:
                    threshold_mask[id] = 1
                final_id = _judge_end(threshold_mask)
                if final_id:
                    print("the best threshold：{}".format(threshold[final_id]))
                    new_graph, memory_planer, is_linearint = _get_memory_plan(
                        graph, platform, local_mem, threshold[final_id], is_dump
                    )
                    break

    print("set linearint threshold and analyze memory end")

    return new_graph, memory_planer


__all__ = ["graph_adapter"]



from typing import List

from ..graph import Graph
from .op_fusion import op_fusion
from .layout_convert import layout_optimizer


def graph_optimizer(graph: Graph, strategy: str, is_dump: bool = False) -> Graph:
    graph = op_fusion(graph, strategy, is_dump)
    graph.init_tensor()
    print("---- op fusion success ----")

    graph = layout_optimizer(graph, is_dump)
    graph.init_tensor()
    print("---- convert layout success ----")
    return graph


__all__ = ["graph_optimizer"]



import math
import numpy as np

from .._type._ctype import tffi
from ...graph import Tensor
from ...enum_defines import DevType
from .base import Operator, OperatorAttrs, register_op


class GRUIntAttrs(OperatorAttrs):
    def checkparams(self) -> None:
        assert "scale_i" in self.attrs
        assert "scale_iw" in self.attrs
        assert "scale_hw" in self.attrs
        assert "scale_h" in self.attrs
        assert "scale_o" in self.attrs
        assert "batch_first" in self.attrs
        assert "hidden_size" in self.attrs

    def serialize(self) -> bytes:
        attrs = tffi.new("GRUIntAttrs *")

        attrs.hidden_size = self.attrs["hidden_size"]
        attrs.input_size = self.attrs["input_size"]
        attrs.layout = self.attrs.get("batch_first", 0)

        return bytes(tffi.buffer(attrs))


@register_op
class GRUInt(Operator):
    def __init__(self, attrs={}):
        self.attrs = GRUIntAttrs(attrs)
        self.weight_index = 1

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) > 3
        X = inputs[0]  # input scale
        i2h_w = inputs[self.weight_index]  ##i2h_w scale
        h2h_w = inputs[self.weight_index + 1]
        assert X.dtype == i2h_w.dtype and X.dtype == h2h_w.dtype
        assert len(X.shape) == 3
        assert len(i2h_w.shape) == 3 or len(i2h_w.shape) == 2
        assert len(h2h_w.shape) == 3 or len(h2h_w.shape) == 2

        scale_i = self.attrs.get("scale_i")
        temp = math.log(scale_i, 2)
        abs(temp - int(temp)) < 0.000001
        assert self.inputs[0].scale == int(temp)

        scale_iw = self.attrs.get("scale_iw")
        temp = math.log(scale_iw, 2)
        abs(temp - int(temp)) < 0.000001
        self.inputs[1].scale = temp

        scale_hw = self.attrs.get("scale_hw")
        temp = math.log(scale_hw, 2)
        abs(temp - int(temp)) < 0.000001
        self.inputs[2].scale = temp

        scale_h = self.attrs.get("scale_h")
        temp4 = math.log(scale_h, 2)
        assert temp4 == int(temp4)

        layout = self.attrs.get("batch_first")
        if layout == 0:
            T = X.shape[0]
            B = X.shape[1]
        else:
            B = X.shape[0]
            T = X.shape[1]
        hidden_size = self.attrs["hidden_size"]
        if layout == 0:
            yshape = list([T, B, hidden_size])
        else:
            yshape = list([B, T, hidden_size])

        scale_o = self.attrs.get("scale_o")
        temp = math.log(scale_o, 2)
        abs(temp - int(temp)) < 0.000001
        Y = Tensor(shape=yshape, dtype=X.dtype, scale=temp)
        self.outputs = [Y]

        hshape = list([1, B, hidden_size])
        hidden_o = Tensor(shape=hshape, dtype=np.float32, scale=temp4)
        self.outputs.append(hidden_o)

    def get_workspace(self, dev_type):
        h2h_weight = self.inputs[self.weight_index + 1]
        hidden_size = h2h_weight.shape[1]
        max_workspace = Tensor.from_shape(
            [hidden_size * 4 * 3 ],
            np.int8,
            dev_type,
        )
        return [max_workspace]

    def pack_params(self, dev_type: DevType):
        weight_i = self.inputs[1]
        weight_h = self.inputs[2]
        data_i = weight_i.data.transpose(1, 0)
        data_h = weight_h.data.transpose(1, 0)
        self.inputs[1].update(data=data_i, shape=data_i.shape)
        self.inputs[2].update(data=data_h, shape=data_h.shape)


__all__ = ["GRUInt"]



from .._type._ctype import tffi

from .base import Operator, OperatorAttrs, register_op
from ...enum_defines import DevType


class SqueezeAttrs(OperatorAttrs):
    def serialize(self) -> bytes:
        attrs = tffi.new("SqueezeAttrs *")
        attrs.axes = self.attrs.get("axes", [])
        attrs.ndim = len(attrs.axes)
        return bytes(tffi.buffer(attrs))


@register_op
class Squeeze(Operator):
    def __init__(self, attrs={}):
        self.attrs = SqueezeAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) == 1
        X = self.inputs[0]
        tShape = list(X.shape)
        axes = self.attrs.get("axes", [])
        if len(axes) == 0:
            yshape = []
            for i, s in enumerate(tShape):
                if s != 1:
                    yshape.append(s)
                else:
                    axes.append(i)
                if len(yshape) == 1:
                    yshape = [1, yshape[0]]                    
        else:
            for x in axes:
                assert x < len(tShape) and x >= -len(tShape)
                assert tShape[x] == 1, "invalid config for squeeze in axis:{}".format(x)
                tShape[x] = None
            yshape = [x for x in tShape if x is not None]
            if len(yshape) == 1:
                yshape = [1, yshape[0]]
                
        Y = X.clone(shape=tuple(yshape))
        self.outputs = [Y]

    def is_inplace(self):
        return True


__all__ = ["Squeeze"]



from ._ctype import tffi
from ._crc24 import crc24

from ._tIO import *
from ._tDebug import *
from ._tDMA import *
from ._tModel import *
from ._tMemory import *
from ._tTensor import *
from ._tOperator import *
from ._tParamter import *
from ...enum_defines import ALIGN16


class tModel(object):
    def __init__(
        self,
        memory: tMemoryList,
        tensor: tTensorList,
        op: tOperatorList,
        io: tIOInfo,
        param: tParameterList,
        debug: tDebugList,
        dma: tDMAList,
    ):

        self.hdr = tffi.new("tModelHeader *")

        self.hdr.label_ = b"thinker10"
        self.hdr.crc32_ = 0

        pack_buff = bytes(tffi.buffer(self.hdr))
        offset = ALIGN16(len(pack_buff))
        head_size = offset
        pack_buff += b"\0" * (offset - len(pack_buff))

        self.hdr.memory_offset_ = offset
        pack_buff += memory.to_bytes()
        offset = ALIGN16(len(pack_buff))
        pack_buff += b"\0" * (offset - len(pack_buff))

        self.hdr.tensor_offset_ = offset
        pack_buff += tensor.to_bytes()
        offset = ALIGN16(len(pack_buff))
        pack_buff += b"\0" * (offset - len(pack_buff))

        self.hdr.op_offset_ = offset
        pack_buff += op.to_bytes()
        offset = ALIGN16(len(pack_buff))
        pack_buff += b"\0" * (offset - len(pack_buff))

        self.hdr.io_offset_ = offset
        pack_buff += io.to_bytes()
        offset = ALIGN16(len(pack_buff))
        pack_buff += b"\0" * (offset - len(pack_buff))

        self.hdr.param_offset_ = offset
        pack_buff += param.to_bytes()
        offset = ALIGN16(len(pack_buff))
        pack_buff += b"\0" * (offset - len(pack_buff))

        self.hdr.debug_offset_ = offset
        pack_buff += debug.to_bytes()
        offset = ALIGN16(len(pack_buff))
        pack_buff += b"\0" * (offset - len(pack_buff))

        self.hdr.dma_offset_ = offset
        pack_buff += dma.to_bytes()
        offset = ALIGN16(len(pack_buff))
        pack_buff += b"\0" * (offset - len(pack_buff))

        crc = crc24(pack_buff[head_size:])

        self.hdr.total_size_ = offset
        self.hdr.crc32_ = crc
        hdr_buff = bytes(tffi.buffer(self.hdr))

        self.bytes = hdr_buff + pack_buff[len(hdr_buff) :]

        print("resource_total_size:{}".format(len(self.bytes)))

    def to_bytes(self):
        return self.bytes


__all__ = ["tModel"]



from ...graph import *
from .._type._ctype import tffi
from ...enum_defines import Layout
from .base import Operator, OperatorAttrs, register_op


class SplitAttrs(OperatorAttrs):
    def checkparams(self) -> None:
        assert "axis" in self.attrs
        assert "dims" in self.attrs

    def serialize(self) -> bytes:
        attrs = tffi.new("SplitAttrs *")

        attrs.axis = self.attrs["axis"]
        attrs.dims = self.attrs["dims"]
        attrs.split = self.attrs["split"]

        return bytes(tffi.buffer(attrs))


@register_op
class Split(Operator):
    def __init__(self, attrs={}):
        self.attrs = SplitAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) == 1
        X = inputs[0]
        axis = self.attrs["axis"]
        assert axis < len(X.shape)

        if self.attrs.get("split", None) == None:
            dims = self.attrs["dims"]
            self.attrs["split"] = []
            for i in range(dims):
                self.attrs["split"].append(X.shape[axis] // dims)

        split = self.attrs["split"]
        assert sum(split) == X.shape[axis]

        outputs = []
        for idx in split:
            shape = list(X.shape)
            shape[axis] = int(idx)
            yi = X.clone(shape=shape)
            outputs.append(yi)
        self.outputs = outputs

    def sub_layout_convert(self):
        inputs = self.inputs
        if inputs[0].layout == Layout.NHWC:
            axis = self.attrs["axis"]
            if axis == 1:
                axis = 3
            elif axis == 2:
                axis = 1
            elif axis == 3:
                axis = 2
            self.attrs["axis"] = axis


__all__ = ["Split"]



import numpy as np
from .Tensor import Tensor
from ..enum_defines import TensorType


class GraphEntry(object):
    def __init__(
        self,
        name="",
        tensor=None,
        is_constant_tensor=False,
        tensor_type=TensorType.Normal,
    ):
        self.tensor = Tensor(shape=(), dtype=np.float32) if tensor is None else tensor
        self.name = name
        self.is_constant_tensor = is_constant_tensor
        self.tensor_type = tensor_type
        self.dst_nodes = []
        self.src_node = None
        self.index = -1

    @property
    def layout(self):
        return self.tensor.layout

    @layout.setter
    def layout(self, value):
        self.tensor.layout = value

    @property
    def data(self):
        return self.tensor.data

    @data.setter
    def data(self, value):
        self.tensor.data = value

    def is_constant(self):
        return self.is_constant_tensor

    def set2_constant(self):
        if not self.tensor.has_data():
            raise ValueError("Can't set tensor_type to constant for empty data entry.")
        self.is_constant_tensor = True

    def is_graph_input(self):
        return self.is_constant_tensor == False and self.tensor_type == TensorType.Input

    def set2_graph_input(self):
        self.is_constant_tensor = False
        self.tensor_type = TensorType.Input

    def is_graph_output(self):
        return (
            self.is_constant_tensor == False and self.tensor_type == TensorType.Output
        )

    def set2_graph_output(self):
        self.is_constant_tensor = False
        self.tensor_type = TensorType.Output

    def set2_graph_normal(self):
        self.is_constant_tensor = False
        self.tensor_type = TensorType.Normal

    def set2_emptry(self):
        self.tensor.data = np.array([])
        self.tensor.shape = ()
        self.is_constant_tensor = True
        self.tensor_type = TensorType.Emptry

    def is_emptry(self):
        return self.is_constant_tensor == True and self.tensor_type == TensorType.Emptry

    def add_dst_node(self, node):
        self.dst_nodes.append(node)

    def copy(self, src: "GraphEntry") -> "GraphEntry":
        self.name = src.name
        self.tensor = src.tensor.clone()
        if isinstance(src.tensor.data, np.ndarray):
            self.tensor.data = np.array(src.tensor.data)
        else:
            self.tensor.data = src.tensor.data
        self.is_constant_tensor = src.is_constant_tensor
        self.tensor_type = src.tensor_type
        self.index = src.index
        return self

    def clone(self) -> "GraphEntry":
        return GraphEntry().copy(self)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        if isinstance(other, GraphEntry):
            return self.name == other.name
        return False

    def __str__(self):
        return "name={}, tensor={}, is_constant={}".format(
            self.name, self.tensor, self.is_constant_tensor
        )

    def __repr__(self):
        return self.__str__()


__all__ = ["GraphEntry"]



from ._ctype import *
from ._tMemory import *
from ._tTensor import *
from ._tOperator import *
from ._tIO import *
from ._tParamter import *
from ._tModel import *
from ._tShapeInfer import *
from ._tShapeOnnxInfer import *
from ._tDebug import *
from ._tDMA import *



from typing import List
from ..enum_defines import Layout
from ..save_model import save_to_onnx_model
from ..graph import Graph, GraphNode, GraphEntry
from ..resource_packer.ops.base import LayoutPerfData
from ..resource_packer.ops.base import create_operator

VARIABLE_TAG = "@"


def _make_layout_name(name: str, layout: Layout) -> str:
    name = name.split(VARIABLE_TAG)[0]
    if layout == Layout.NCHW:
        return name
    elif layout == Layout.NHWC:
        return name + VARIABLE_TAG + "NHWC"
    elif layout == Layout.NCWH:
        return name + VARIABLE_TAG + "NCWH"


def _init_graph(src_graph: Graph) -> List[Graph]:
    graph = Graph()
    graph.copy_attrs(src_graph)

    for entry in src_graph.inputs:
        new_entry = entry.clone()
        graph.add_entry(new_entry)
        graph.inputs.append(new_entry)

    for entry in src_graph.entries.values():
        if entry.is_constant():
            new_entry = entry.clone()
            graph.add_entry(new_entry)
    return [graph]


def _beam_search(graph_list: List[Graph], beam_size: int) -> List[Graph]:
    if len(graph_list) <= beam_size:
        return graph_list
    perfs = [graph.performance for graph in graph_list]
    import heapq

    indexs = heapq.nsmallest(beam_size, range(len(perfs)), perfs.__getitem__)
    dst_graph = [graph_list[index] for index in indexs]
    return dst_graph


def _add_pack_node(
    graph: Graph, src_name: str, src_layout: Layout, dst_name: str, dst_layout: Layout
) -> None:
    src_entry = graph.entries[src_name]
    dst_entry = src_entry.clone()
    dst_entry.name = dst_name
    dst_entry.set2_graph_normal()
    dst_entry.layout = dst_layout
    graph.add_entry(dst_entry)

    pack_node = GraphNode("Packing", dst_name + "_packing")
    pack_node.inputs = [src_entry]
    pack_node.outputs = [dst_entry]
    pack_node.op = create_operator(
        "Packing", {}, [src_entry.tensor], [dst_entry.tensor]
    )
    permute_perfs = pack_node.op.get_layout_perf()
    for permute_perf in permute_perfs:
        if permute_perf.inputs_layout[0] == src_layout:
            graph.performance += permute_perf.performance
            break
    graph.add_node(pack_node)


def _get_entry_by_name(graph: Graph, src_name: str) -> GraphEntry:
    entry = None
    name = src_name.split(VARIABLE_TAG)[0]
    for e in graph.entries.values():
        e_name = e.name.split(VARIABLE_TAG)[0]
        if e_name == name and e.name != src_name:
            entry = e
            break
    if entry == None:
        raise ("not find input")
    return entry


def _add_node_to_graph(graph: Graph, node: GraphNode, perf: LayoutPerfData) -> None:
    new_node = node.clone()
    for i, entry in enumerate(new_node.inputs):
        # not params
        if entry.is_constant():
            continue
        entry.layout = perf.inputs_layout[i]
        entry.name = _make_layout_name(entry.name, entry.layout)

        if entry.name not in graph.entries:
            # sample, need nchw layout,but not exist in graph, then add packing node
            temp_entry = _get_entry_by_name(graph, entry.name)
            _add_pack_node(
                graph, temp_entry.name, temp_entry.layout, entry.name, entry.layout
            )
        new_node.inputs[i] = graph.entries[entry.name]

    for i, entry in enumerate(new_node.outputs):
        entry.layout = perf.outputs_layout[i]
        entry.name = _make_layout_name(entry.name, entry.layout)
        graph.add_entry(entry)
        new_node.outputs[i] = graph.entries[entry.name]

        # output node must be nchw layout, add packing node
        if entry.is_graph_output() and entry.layout != Layout.NCHW:
            entry.set2_graph_normal()
            output_layout = Layout.NCHW
            output_name = _make_layout_name(entry.name, output_layout)
            _add_pack_node(graph, entry.name, entry.layout, output_name, output_layout)
            graph.entries[output_name].set2_graph_output()

    # add node and performance
    graph.add_node(new_node)
    graph.performance += perf.performance


def _packing2transpose(node: GraphNode) -> None:
    in_layout = node.inputs[0].layout
    out_layout = node.outputs[0].layout
    if in_layout == Layout.NCHW and out_layout == Layout.NHWC:
        node.op_type = "Transpose"
        node.attrs["perm"] = [0, 2, 3, 1]
        node.op = create_operator(
            "Transpose", node.attrs, [node.inputs[0].tensor], [node.outputs[0].tensor]
        )

    elif in_layout == Layout.NHWC and out_layout == Layout.NCHW:
        node.op_type = "Transpose"
        node.attrs["perm"] = [0, 3, 1, 2]
        node.op = create_operator(
            "Transpose", node.attrs, [node.inputs[0].tensor], [node.outputs[0].tensor]
        )
    elif (in_layout == Layout.NCHW and out_layout == Layout.NCWH) or (
        in_layout == Layout.NCWH and out_layout == Layout.NCHW
    ):
        node.op_type = "Transpose"
        node.attrs["perm"] = [0, 1, 3, 2]
        node.op = create_operator(
            "Transpose", node.attrs, [node.inputs[0].tensor], [node.outputs[0].tensor]
        )


def _post_process_graph(graph: Graph) -> None:
    for node in graph.nodes.values():
        if node.op_type == "Packing":
            _packing2transpose(node)
        # change shape
        node.op.layout_convert()


# layout optimizer by device
def _layout_convert(src_graph: Graph) -> Graph:
    beam_size = 2
    beam_search_graphs = _init_graph(src_graph)
    for node in src_graph.nodes.values():
        if node.op == None:
            raise ("node op must be not None")
        perf_list = node.op.get_layout_perf()

        search_graphs_expand = []
        for perf in perf_list:
            for graph in beam_search_graphs:
                new_graph = graph.clone()
                _add_node_to_graph(new_graph, node, perf)
                search_graphs_expand.append(new_graph)
        beam_search_graphs = _beam_search(search_graphs_expand, beam_size)

    for graph in beam_search_graphs:
        for output in src_graph.outputs:
            graph.entries[output.name].set2_graph_output()
            graph.outputs.append(graph.entries[output.name])
    best_graph = _beam_search(beam_search_graphs, 1)[0]
    best_graph.update()
    _post_process_graph(best_graph)
    return best_graph


def layout_optimizer(
    graph: Graph,
    is_dump: bool = True,
    dump_file_path: str = "./model.ignore/graph_layout_convert.onnx",
) -> Graph:
    """
    layout转换策略：在conv层的输入数据大于64KB，且W>H时，交换W和H的维度
    """
    new_graph = _layout_convert(graph)
    if is_dump:
        save_to_onnx_model(new_graph, dump_file_path)
    return new_graph


__all__ = ["layout_optimizer"]



import numpy as np
from ctypes import Structure, sizeof
from ctypes import c_uint16, c_uint32

from ...enum_defines import ALIGN16


class tIO(object):
    def __init__(self, tensor_id, name):
        self.tensor_id = tensor_id
        self.name = name


class tState(object):
    def __init__(self, in_tensor_id, out_tensor_id, name):
        self.in_tensor_id = in_tensor_id
        self.out_tensor_id = out_tensor_id
        self.name = name


class tIOInfo(Structure):
    _fields_ = [
        ("num_input", c_uint16),
        ("num_output", c_uint16),
        ("num_state", c_uint16),
        ("name_length", c_uint16),
        ("tensor_offset", c_uint32),
        ("name_offset", c_uint32),
    ]

    def __init__(self, inputs, outputs, states=[]):
        self.num_input = c_uint16(len(inputs))
        self.num_output = c_uint16(len(outputs))
        self.num_state = c_uint16(len(states))

        self.tensor_ids = (
            [x.tensor_id for x in inputs]
            + [x.tensor_id for x in outputs]
            + [x.in_tensor_id for x in states]
            + [x.out_tensor_id for x in states]
        )

        self.name_list = (
            [x.name for x in inputs]
            + [x.name for x in outputs]
            + [x.name for x in states]
        )

        max_len = max([len(x) for x in self.name_list])
        max_len = ALIGN16(max_len + 1)
        self.name_length = c_uint16(max_len)

        self.tensor_offset = sizeof(self)
        tensor_bytes = np.array(self.tensor_ids, dtype=np.int32).tobytes()
        self.name_offset = self.tensor_offset + len(tensor_bytes)

        self.bytes = bytes(self)
        self.bytes += tensor_bytes
        for name in self.name_list:
            self.bytes += np.array([ord(x) for x in name], dtype=np.uint8).tobytes()
            self.bytes += b"\0" * (max_len - len(name))

    def to_bytes(self):
        return self.bytes


__all__ = ["tIO", "tIOInfo"]



from typing import Any, Dict, List, Optional

from .._type._ctype import tffi
from ...enum_defines import DevType, Layout
from .base import Operator, OperatorAttrs, register_op


class TransposeAttrs(OperatorAttrs):
    def __init__(self, attrs: Optional[Dict[str, Any]] = {}) -> None:
        super().__init__(attrs, "TransposeAttrs")

    def checkparams(self) -> None:
        self.attrs["perm"] = self.attrs.get("perm", [])
        assert len(self.attrs["perm"]) < 7
        self.attrs["ndim"] = len(self.attrs["perm"])

    def serialize(self) -> bytes:
        attrs = tffi.new("TransposeAttrs *")

        attrs.axes_ = self.attrs["perm"]
        attrs.ndim_ = self.attrs["ndim"]

        return bytes(tffi.buffer(attrs))


@register_op
class Transpose(Operator):
    def __init__(self, attrs={}):
        self.attrs = TransposeAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) == 1, "Check fail: number of input not equal 1"
        X = inputs[0]
        shape = list(X.shape)
        axes = self.attrs["perm"]
        if len(axes) == 0:
            axes = list(range(len(shape) - 1, -1, -1))
        assert len(axes) <= len(shape)
        new_shape = [shape[x] for x in axes]
        new_shape += shape[len(axes) :]

        Y = X.clone(shape=tuple(new_shape))
        if X.layout == Layout.NCHW and tuple(axes) == (0, 1, 3, 2):
            Y.layout = Layout.NCWH
        elif X.layout == Layout.NCWH and tuple(axes) == (0, 1, 3, 2):
            Y.layout = Layout.NCHW

        self.outputs = [Y]

__all__ = ["Transpose"]



import math
import numpy as np
from enum import Enum

from ...graph import Tensor
from .._type._ctype import tffi
from ...enum_defines import DevType, Layout
from .utils import get_deconv_workspace_desc
from .base import Operator, OperatorAttrs, register_op
from .utils import attr2tuple, calc_deconv2d_output_shape


class ConvTranspose2dIntAttrs(OperatorAttrs):
    def checkparams(self) -> None:
        attrs = self.attrs
        if "auto_pad" in attrs and attrs["auto_pad"] != "NOTSET":
            auto_pad = attrs["auto_pad"]
            raise (f"[ConvTranspose2dInt] Not supported yet: auto_pad:{auto_pad}.")
        if "output_shape" in attrs and tuple(attrs["output_shape"]) != (0, 0):
            output_shape = tuple(attrs["output_shape"])
            raise (
                f"[ConvTranspose2dInt] Not supported yet: outputshape:{output_shape}."
            )
        if (
            "output_padding" in attrs
            and tuple(attrs["output_padding"]) != (0, 0)
            and tuple(attrs["output_padding"]) != (0, 0, 0, 0)
            and tuple(attrs["output_padding"]) != (0, 0, 0, 0, 0, 0)
        ):
            output_padding = tuple(attrs["output_padding"])
            raise (
                f"[ConvTranspose2dInt] Not supported yet: output_padding:{output_padding}."
            )
        else:
            output_padding = (0, 0, 0, 0, 0, 0)

        dilations = attrs.get("dilations")
        dilations = attr2tuple(dilations, (1, 1))
        assert dilations == (1, 1), "not support dilation ConvTranspose2dInt!"

        kernels = attrs.get("kernel_shape")
        kernels = attr2tuple(kernels, (1, 1))
        assert kernels[0] in {
            1,
            2,
            3,
            4,
            5,
        }, "kernel_w for ConvTranspose2dInt exceed limit"
        assert kernels[1] in {
            1,
            2,
            3,
            4,
            5,
        }, "kernel_h for ConvTranspose2dInt exceed limit"

        pads = attrs.get("pads")
        pads = attr2tuple(pads, (0, 0, 0, 0))
        assert pads[0] in {
            0,
            1,
            2,
            3,
            4,
        }, "pad_left for ConvTranspose2dInt exceed limit"
        assert pads[1] in {
            0,
            1,
            2,
            3,
            4,
        }, "pad_right for ConvTranspose2dInt exceed limit"
        assert pads[2] in {0, 1, 2, 3, 4}, "pad_up for ConvTranspose2dInt exceed limit"
        assert pads[3] in {
            0,
            1,
            2,
            3,
            4,
        }, "pad_down for ConvTranspose2dInt exceed limit"

        strides = attrs.get("strides")
        strides = attr2tuple(strides, (1, 1))
        assert strides[0] in {1, 2, 4}, "stride_h for ConvTranspose2dInt exceed limit"
        assert strides[1] in {1, 2, 4}, "stride_w for ConvTranspose2dInt exceed limit"

        if strides[0] == 2:
            assert kernels[0] in {2, 3, 4, 5}, "stride_h and kernel_h do not match"
        if strides[0] == 4:
            assert kernels[0] in {4, 5}, "stride_h and kernel_h do not match"
        if strides[1] == 2:
            assert kernels[1] in {2, 3, 4, 5}, "stride_w and kernel_w do not match"
        if strides[1] == 4:
            assert kernels[1] in {4, 5}, "stride_w and kernel_w do not match"

        group = int(attrs.get("group", 1))

        if isinstance(attrs.get("layout", "NCHW"), Enum):
            layout = attrs.get("layout", "NCHW")
        else:
            layout = Layout.from_str(attrs.get("layout", "NCHW"))
        assert layout in {Layout.NCHW, Layout.NHWC}

    def serialize(self) -> bytes:
        attrs = tffi.new("Conv2dIntAttrs *")

        attrs.dilation = self.attrs["dilations"]
        attrs.kernel = self.attrs["kernel_shape"]
        attrs.pad = self.attrs["pads"]
        attrs.stride = self.attrs["strides"]
        attrs.group = self.attrs["group"]
        attrs.layout = self.attrs["layout"].value
        attrs.quant_type = self.attrs["quant_type"].value
        attrs.act_type = self.attrs["act_type"]

        return bytes(tffi.buffer(attrs))


@register_op
class ConvTranspose2dInt(Operator):
    def __init__(self, attrs={}):
        self.attrs = ConvTranspose2dIntAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) in {2, 3}
        X = inputs[0]
        W = inputs[1]
        assert len(X.shape) == 4 and len(W.shape) == 4

        scale_x = self.attrs.get("scale_x")
        temp = math.log(scale_x, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert X.scale == int(
            temp
        ), "scale of tensor must be same with scale_x in attribute"

        scale_w = self.attrs.get("scale_w")
        temp = math.log(scale_w, 2)
        assert abs(temp - int(temp)) < 0.000001
        self.inputs[1].scale = int(temp)

        scale_o = self.attrs.get("scale_o")
        temp = math.log(scale_o, 2)
        assert abs(temp - int(temp)) < 0.000001

        shape = calc_deconv2d_output_shape(
            input_shape=X.shape,
            weight_shape=W.shape,
            kernels=self.attrs["kernel_shape"],
            strides=self.attrs["strides"],
            dilations=self.attrs["dilations"],
            pads=self.attrs["pads"],
            groups=self.attrs["group"],
            layout=Layout.NCHW,
        )
        Y = X.clone(shape=tuple(shape), scale=int(temp))
        self.outputs = [Y]
        return self.outputs

    def get_workspace(self, dev_type):
        workspace_size = get_deconv_workspace_desc(
            None, self.inputs[0], self.inputs[1], self.attrs, self.outputs
        )
        max_workspace = Tensor.from_shape([workspace_size], np.int8, dev_type)
        return [max_workspace]

    def pack_params(self, dev_type: DevType):
        if dev_type == DevType.LUNA or dev_type == DevType.HIFI:
            kernels = self.attrs["kernel_shape"]
            kernel_h = kernels[0]
            kernel_w = kernels[1]
            assert 1 <= kernel_w <= 5
            assert 1 <= kernel_h <= 5

            dilations = self.attrs["dilations"]
            dilation_w = dilations[0]
            dilation_h = dilations[1]
            strides = self.attrs["strides"]
            stride_w = strides[0]
            stride_h = strides[1]
            assert stride_w == 1 or stride_w == 2 or stride_w == 4
            assert stride_h == 1 or stride_h == 2 or stride_h == 4
            group = self.attrs["group"]
            pads = self.attrs["pads"]
            assert 0 <= pads[0] <= 4
            assert 0 <= pads[1] <= 4
            assert 0 <= pads[2] <= 4
            assert 0 <= pads[3] <= 4

            if len(self.inputs) >= 2:
                bias = self.inputs[2]
                if bias.dtype != np.dtype(np.int32):
                    new_bias = np.zeros(bias.shape, np.dtype(np.int32))
                    for i in range(bias.shape[0]):
                        new_bias[i] = bias.data[i]
                    self.inputs[2].update(data=new_bias, dtype=np.dtype(np.int32))

            weight_data = self.inputs[1]
            if weight_data.layout == Layout.NHWC8:
                return
            else:
                kernel_num = weight_data.shape[1]
                kernel_c = weight_data.shape[0]
                assert kernel_c * group == self.inputs[0].shape[1]
                assert kernel_h == weight_data.shape[2]
                assert kernel_w == weight_data.shape[3]

                h = self.inputs[0].shape[2]
                w = self.inputs[0].shape[3]
                data_size = (
                    ((w + 8 * stride_w - 1) // (8 * stride_w))
                    * (8 * stride_w)
                    * ((kernel_c + 7) // 8)
                    * 8
                    * h
                )
                assert data_size * self.inputs[0].dtype.itemsize <= 65536

                if group != 1:
                    assert kernel_num % group == 0
                    num_input_align = ((kernel_c // group + 7) // 8) * 8
                    num_output_align = ((kernel_num + 1) // 2) * 2
                    kernel_size = (
                        num_input_align * num_output_align * kernel_h * kernel_w
                    )
                    assert kernel_size * weight_data.dtype.itemsize <= 32768
                    new_weight_data = np.zeros(
                        (group, num_output_align, num_input_align, kernel_h, kernel_w),
                        weight_data.dtype,
                    )
                    shape = (
                        group * num_output_align,
                        kernel_h,
                        kernel_w,
                        num_input_align,
                    )
                    weight_data_split = weight_data.data.reshape(
                        group, -1, kernel_num, kernel_h, kernel_w
                    )
                    weight_data_transpose = weight_data_split.transpose(0, 2, 1, 3, 4)
                    # weight_data.data = weight_data.data.reshape(group, -1, kernel_c, kernel_h, kernel_w)

                    for g in range(group):
                        for p in range(kernel_num):
                            for q in range(kernel_c // group):
                                new_weight_data[g, p, q, :, :] = weight_data_transpose[
                                    g, p, q, :, :
                                ]
                    new_weight_data = np.flip(new_weight_data, (3, 4))

                    new_weight_data = new_weight_data.transpose(0, 1, 3, 4, 2)
                    # shape = new_weight_data.shape
                    temp = new_weight_data.reshape(
                        group, -1, 2, kernel_h, kernel_w, num_input_align
                    )
                    temp1 = temp.transpose(0, 1, 3, 4, 2, 5)
                    shape1 = temp1.shape
                    temp2 = temp1.reshape(
                        shape1[0], shape1[1], shape1[2], shape1[3], shape1[4], -1, 8
                    )
                    new_weight_data = temp2.transpose(0, 1, 2, 3, 5, 4, 6)
                    self.inputs[1].update(
                        shape=shape, data=new_weight_data, layout=Layout.NHWC8
                    )
                else:
                    num_input_align = ALIGN8(kernel_c)
                    num_output_align = ((kernel_num + 1) // 2) * 2
                    kernel_size = (
                        num_input_align * num_output_align * kernel_h * kernel_w
                    )
                    assert kernel_size * weight_data.dtype.itemsize <= 32768

                    new_weight_data = np.zeros(
                        (num_output_align, num_input_align, kernel_h, kernel_w),
                        weight_data.dtype,
                    )
                    weight_data_tranpose = weight_data.data.transpose(1, 0, 2, 3)

                    for p in range(kernel_num):
                        for q in range(kernel_c):
                            new_weight_data[p, q, :, :] = weight_data_tranpose[
                                p, q, :, :
                            ]
                    new_weight_data = np.flip(new_weight_data, (2, 3))
                    # new_weight_data = np.flip(new_weight_data, 3)

                    new_weight_data = new_weight_data.transpose(0, 2, 3, 1)
                    shape = new_weight_data.shape
                    temp = new_weight_data.reshape(
                        -1, 2, kernel_h, kernel_w, num_input_align
                    )
                    new_weight_data = temp.transpose(0, 2, 3, 1, 4)
                    shape1 = new_weight_data.shape
                    new_weight_data = new_weight_data.reshape(
                        shape1[0], shape1[1], shape1[2], shape1[3], -1, 8
                    )
                    new_weight_data = new_weight_data.transpose(0, 1, 2, 4, 3, 5)
                    self.inputs[1].update(
                        shape=shape, data=new_weight_data, layout=Layout.NHWC8
                    )
                    return


__all__ = ["ConvTranspose2dInt"]



from .GraphEntry import GraphEntry


class GraphNode(object):
    def __init__(self, op, name, attrs={}, inputs=[], outputs=[]):
        self.name = name  # name with scope
        self.op_type = op
        self.dev_type = None
        self.attrs = attrs.copy()
        self.inputs: list[GraphEntry] = [x for x in inputs]
        self.outputs: list[GraphEntry] = [x for x in outputs]
        self.index = -1
        self.op = None

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        if isinstance(other, GraphNode):
            return self.name == other.name
        return False

    def __str__(self):
        return "{}('{}'): attrs={}".format(self.op_type, self.name, self.attrs)

    def __repr__(self):
        return str(self)

    def ref_input_nodes(self):
        nodes = []
        for entry in self.inputs:
            n = entry.src_node
            if n is not None:
                nodes.append(n)
        return nodes

    def ref_output_nodes(self):
        nodes = []
        for entry in self.outputs:
            for n in entry.dst_nodes:
                nodes.append(n)
        return nodes

    def get_input_index(self, entry: GraphEntry):
        return self.inputs.index(entry)

    def get_output_index(self, entry: GraphEntry):
        return self.outputs.index(entry)

    def copy(self, src: "GraphNode") -> "GraphNode":
        self.name = src.name
        self.op_type = src.op_type
        self.dev_type = src.dev_type
        self.attrs = src.attrs.copy()
        self.outputs = [output.clone() for output in src.outputs]
        self.inputs = [input.clone() for input in src.inputs]
        self.index = src.index
        self.op = src.op
        return self

    def clone(self) -> "GraphNode":
        return GraphNode(None, None).copy(self)

    def create_op(self) -> None:
        inputs = [input.tensor for input in self.inputs]
        outputs = [output.tensor for output in self.outputs]
        from ..resource_packer.ops.base import create_operator

        self.op = create_operator(self.op_type, self.attrs, inputs, outputs)


__all__ = ["GraphNode"]



from ._ctype import tffi


def nptype_to_thinker(nptype):
    import numpy as np

    nptype = np.dtype(nptype)
    dtype = (ord(nptype.str[-2]) << 8) + int(nptype.str[-1])
    return dtype


class tTensor(object):
    def __init__(self, tensor, mem_id, mem_offset, mem_type):
        dtype = nptype_to_thinker(tensor.dtype)

        self.obj = tffi.new("tTensor *")
        self.obj.mem_.type_ = mem_type
        self.obj.dtype_ = dtype
        self.obj.mem_id_ = mem_id
        self.obj.scale_ = tensor.scale
        self.obj.zero_ = tensor.zero
        self.obj.shape_.ndim_ = len(tensor.shape)
        self.obj.shape_.dims_ = tensor.shape
        self.obj.offset_ = mem_offset
        self.obj.layout_ = tensor.layout.value

    def to_bytes(self):
        return bytes(tffi.buffer(self.obj))


class tTensorList(object):
    def __init__(self, tensor_list):
        self._list = tensor_list
        self.obj = tffi.new("tTensorList *")
        self.obj.count_ = len(tensor_list)
        self.obj.elem_size_ = tffi.sizeof("tTensor")
        self.obj.header_size_ = tffi.sizeof("tTensorList")
        self.obj.offset_ = tffi.sizeof("tTensorList")

        self.bytes = bytes(tffi.buffer(self.obj))
        for x in self._list:
            self.bytes += x.to_bytes()

    def to_bytes(self):
        return self.bytes


__all__ = ["tTensor", "tTensorList"]



from ...enum_defines import DevType
from .base import Operator, UnaryOperator, BaseLayout, OperatorAttrs, register_op


@register_op
class Relu(UnaryOperator, BaseLayout):
    pass


@register_op
class PRelu(Operator):
    def __init__(self, attrs={}):
        self.attrs = OperatorAttrs(attrs)

    def infer_tensor(self):
        assert len(self.inputs) == 2, "num of inputs in PRelu must be 2"
        X1 = self.inputs[0]
        X2 = self.inputs[1]

        shape1 = list(X1.shape)
        shape2 = list(X2.shape)
        # expand to same dim
        if len(shape1) >= len(shape2):
            shape2 = [1] * (len(shape1) - len(shape2)) + shape2
        else:
            raise AttributeError

        assert len(shape1) == len(shape2)
        shape = [1] * len(shape1)
        for i in range(len(shape1)):
            if shape2[i] == 1:
                shape[i] = shape1[i]
            elif shape1[i] == shape2[i]:
                shape[i] = shape2[i]
            else:
                raise AttributeError
        Y = X1.clone(shape=tuple(shape))
        self.outputs = [Y]


__all__ = ["Relu", "PRelu"]



from typing import List

from ._ctype import tffi
from ...enum_defines import MemType, ALIGN16


class tMemory(object):  # little endian
    def __init__(self, mem_type, size):
        self.obj = tffi.new("tMemory *")
        self.obj.dev_type_ = mem_type
        self.obj.mem_type_ = 0

        size = ALIGN16(size)
        self.obj.size_ = int(size)
        self.obj.dptr_ = 0

    def to_bytes(self):
        return bytes(tffi.buffer(self.obj))


class tMemoryList(object):
    def __init__(
        self, shared_memory_list: List[tMemory], runtime_memory_list: List[tMemory]
    ):
        memory_list = []
        for i in range(len(runtime_memory_list)):
            memory_list += runtime_memory_list[i]
        if shared_memory_list == None:
            self._list = memory_list
        else:
            self._list = shared_memory_list + memory_list

        _all_dev = []
        for i in range(len(self._list)):
            if self._list[i].obj.dev_type_ not in _all_dev:
                _all_dev.append(self._list[i].obj.dev_type_)

        for i in _all_dev:
            _total_size = 0
            for j in range(len(self._list)):
                if i == self._list[j].obj.dev_type_:
                    _total_size += self._list[j].obj.size_
            print("{} need capacity: {} Bytes".format(MemType(i), _total_size))
            if i == 2:
                assert (
                    _total_size < 640 * 1024
                ), "SHARE-MEM to be allocated was {}, exceed 640KB".format(_total_size)
            elif i == 1:
                assert (
                    _total_size < 8 * 1024 * 1024
                ), "PSRAM to be allocated was {}, exceed 8MB".format(_total_size)

        self.obj = tffi.new("tMemoryList *")

        self.obj.shared_count_ = len(shared_memory_list)
        self.obj.total_count_ = len(self._list)
        self.obj.elem_size_ = tffi.sizeof("tMemory")
        self.obj.header_size_ = tffi.sizeof("tMemoryList")
        self.obj.offset_ = ALIGN16(self.obj.header_size_)

        self.bytes = bytes(tffi.buffer(self.obj))
        for x in self._list:
            self.bytes += x.to_bytes()

    def to_bytes(self):
        return self.bytes


__all__ = ["tMemory", "tMemoryList"]



import numpy as np
import math
from .utils import QuantType
from .._type._ctype import tffi
from .base import Operator, OperatorAttrs, register_op


class iqTanhOperatorAttrs(OperatorAttrs):
    def checkparams(self) -> None:
        assert "scale_x" in self.attrs
        assert "scale_o" in self.attrs
        assert "platform_quant" in self.attrs

    def serialize(self) -> bytes:
        attrs = tffi.new("iqBinaryAttrs *")

        quant_type = self.attrs.get("platform_quant")
        attrs.quant_type = QuantType.from_str(quant_type).value

        return bytes(tffi.buffer(attrs))


@register_op
class iqTanh(Operator):
    def __init__(self, attrs={}):
        self.attrs = iqTanhOperatorAttrs(attrs)

    def infer_tensor(self):
        assert len(self.inputs) == 1

        X = self.inputs[0]
        assert X.dtype == np.int16, "input of iqtanh must be int16"

        scale_x = self.attrs.get("scale_x")
        temp = math.log(scale_x, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert X.scale == int(
            temp
        ), "scale of tensor must be same with scale_x in attribute"
        assert X.scale == 11, "scale of iqtanh input must be Q11"

        scale_o = self.attrs.get("scale_o")
        temp = math.log(scale_o, 2)
        assert abs(temp - int(temp)) < 0.000001

        Y = X.clone(scale=int(temp))
        self.outputs = [Y]
        if all([x.has_data() for x in self.inputs]):
            self.forward()


__all__ = ["iqTanh"]



import math

from .utils import AutoPad, CeilMode
from ....enum_defines import Layout


def handle_conv_auto_pad(
    in_dim, stride, kernel, dilation, pad_type, pad_head, pad_tail
):
    if pad_type == AutoPad.NOTSET:
        return pad_head, pad_tail
    elif pad_type == AutoPad.VALID:
        return 0, 0
    elif pad_type in {AutoPad.SAME_LOWER, AutoPad.SAME_UPPER}:
        target_size = (in_dim + stride - 1) // stride
        pad_needed = (target_size - 1) * stride + kernel - in_dim
        if pad_type == AutoPad.SAME_LOWER:
            pad_head = (pad_needed + 1) // 2
        else:
            pad_head = pad_needed // 2
        pad_tail = pad_needed - pad_head
        return pad_head, pad_tail
    else:
        raise (f"Pad type {pad_type} not supported.")


def compute_conv_output_size(in_dim, stride, kernel, dilation, pad_head, pad_tail):
    dkernel = dilation * (kernel - 1) + 1
    output_size = (in_dim + pad_head + pad_tail - dkernel) // stride + 1
    return output_size


def calc_conv_output_dim(
    input_size, kernel, stride, dilation, pad_head, pad_tail, auto_pad
):
    pad_head, pad_tail = handle_conv_auto_pad(
        input_size, stride, kernel, dilation, auto_pad, pad_head, pad_tail
    )
    output_size = compute_conv_output_size(
        input_size, stride, kernel, dilation, pad_head, pad_tail
    )
    return output_size


def calc_conv1d_output_shape(
    input_shape,
    weight_shape,
    kernels=None,
    strides=(1),
    dilations=(1),
    pads=(0, 0),
    groups=1,
    auto_pad=AutoPad.NOTSET,
    layout=Layout.NCHW,
):
    b, c, h = input_shape
    out_c, in_c, k_h = weight_shape
    if layout == Layout.NHWC:
        b, h, c = input_shape
        out_c, k_h, in_c = weight_shape
    assert groups * in_c == c, "Check fail: groups * channel(weight) != channel(input)"
    assert k_h == kernels[0], "Check fail: kernel_h != kernels[0]"
    out_h = calc_conv_output_dim(
        h, kernels[0], strides[0], dilations[0], pads[0], pads[1], auto_pad
    )
    output_shape = (b, out_c, out_h)
    if layout == Layout.NHWC:
        output_shape = (b, out_h, out_c)
    return output_shape


def calc_conv2d_output_shape(
    input_shape,
    weight_shape,
    kernels=None,
    strides=(1, 1),
    dilations=(1, 1),
    pads=(0, 0, 0, 0),
    groups=1,
    auto_pad=AutoPad.NOTSET,
):
    b, c, h, w = input_shape
    out_c, in_c, k_h, k_w = weight_shape
    assert groups * in_c == c, "Check fail: groups * channel(weight) != channel(input)"
    assert (
        k_h == kernels[0] and k_w == kernels[1]
    ), "Check fail: kernel_h != kernels[0] && kernel_w != kernels[1]"
    out_h = calc_conv_output_dim(
        h, kernels[0], strides[0], dilations[0], pads[0], pads[2], auto_pad
    )
    out_w = calc_conv_output_dim(
        w, kernels[1], strides[1], dilations[1], pads[1], pads[3], auto_pad
    )
    output_shape = (b, out_c, out_h, out_w)
    return output_shape


def compute_deconv_total_pad(
    input_size, stride, output_pad, kernel, dilation, output_size
):
    return max(
        0,
        (input_size - 1) * stride
        + output_pad
        + (kernel - 1) * dilation
        + 1
        - output_size,
    )


def handle_deconv_auto_pad(auto_pad, total_pad):
    if auto_pad == AutoPad.SAME_UPPER:
        pad_head = total_pad - total_pad / 2
        pad_tail = total_pad / 2
    else:  # for auto_pad is NOTSET, SAME_LOWER or VALID
        pad_head = total_pad / 2
        pad_tail = total_pad - total_pad / 2
    return pad_head, pad_tail


def calc_deconv_output_dim(
    input_size,
    kernel,
    stride,
    dilation,
    pad_head,
    pad_tail,
    auto_pad,
    output_pad,
    output_size=None,
):
    if output_size != 0:
        assert output_size >= 0
        total_pad = compute_deconv_total_pad(
            input_size, stride, output_pad, kernel, dilation, output_size
        )
        pad_head, pad_tail = handle_deconv_auto_pad(auto_pad, total_pad)
        return output_size
    if auto_pad in {AutoPad.SAME_LOWER, AutoPad.SAME_UPPER}:
        total_pad = compute_deconv_total_pad(
            input_size, stride, output_pad, kernel, dilation, input_size * stride
        )
        pad_head, pad_tail = handle_deconv_auto_pad(auto_pad, total_pad)
    output_size = (
        (input_size - 1) * stride
        + output_pad
        + (kernel - 1) * dilation
        + 1
        - pad_head
        - pad_tail
    )
    return output_size


def calc_deconv2d_output_shape(
    input_shape,
    weight_shape,
    kernels=None,
    strides=(1, 1),
    dilations=(1, 1),
    pads=(0, 0, 0, 0),
    groups=1,
    auto_pad=AutoPad.NOTSET,
    output_pads=(0, 0),
    output_shape=(0, 0),
    layout=Layout.NCHW,
):
    b, c, h, w = input_shape
    in_c, out_c, k_h, k_w = weight_shape
    if layout == Layout.NHWC:
        b, h, w, c = input_shape
        in_c, k_h, k_w, out_c = weight_shape
    if kernels is None:
        kernels = (k_h, k_w)
    assert in_c == c, "Check fail: groups * channel(weight) != channel(input)"
    assert (
        k_h == kernels[0] and k_w == kernels[1]
    ), "Check fail: kernel_h != kernels[0] && kernel_w != kernels[1]"
    out_h = calc_deconv_output_dim(
        h,
        kernels[0],
        strides[0],
        dilations[0],
        pads[0],
        pads[2],
        auto_pad,
        output_pads[0],
        output_shape[0],
    )
    out_w = calc_deconv_output_dim(
        w,
        kernels[1],
        strides[1],
        dilations[1],
        pads[1],
        pads[3],
        auto_pad,
        output_pads[1],
        output_shape[1],
    )
    final_output_shape = (b, groups * out_c, out_h, out_w)
    if layout == Layout.NHWC:
        final_output_shape = (b, groups * out_h, out_w, out_c)
    return final_output_shape


def compute_pool_output_size(
    in_dim, stride, kernel, dilation, pad_head, pad_tail, ceil_mode
):
    dkernel = dilation * (kernel - 1) + 1
    output_size = (in_dim + pad_head + pad_tail - dkernel) / stride + 1
    if ceil_mode == CeilMode.CEIL:
        return math.ceil(output_size)
    return math.floor(output_size)


def calc_pool_output_dim(
    input_size, kernel, stride, dilation, pad_head, pad_tail, auto_pad, ceil_mode
):
    pad_head, pad_tail = handle_conv_auto_pad(
        input_size, stride, kernel, dilation, auto_pad, pad_head, pad_tail
    )
    output_size = compute_pool_output_size(
        input_size, stride, kernel, dilation, pad_head, pad_tail, ceil_mode
    )
    return output_size


def calc_pool2d_output_shape(
    input_shape,
    kernels,
    strides=(1, 1),
    dilations=(1, 1),
    pads=(0, 0, 0, 0),
    auto_pad=AutoPad.NOTSET,
    ceil_mode=CeilMode.NO,
    layout=Layout.NCHW,
):
    b, c, h, w = input_shape
    if layout == Layout.NHWC:
        b, h, w, c = input_shape
    out_h = calc_pool_output_dim(
        h, kernels[0], strides[0], dilations[0], pads[0], pads[2], auto_pad, ceil_mode
    )
    out_w = calc_pool_output_dim(
        w, kernels[1], strides[1], dilations[1], pads[1], pads[3], auto_pad, ceil_mode
    )
    output_shape = (b, c, out_h, out_w)
    if layout == Layout.NHWC:
        output_shape = (b, out_h, out_w, c)
    return output_shape


__all__ = [
    "calc_conv1d_output_shape",
    "calc_conv2d_output_shape",
    "calc_deconv2d_output_shape",
    "calc_pool2d_output_shape",
]



import math
import numpy as np
from typing import Any, Dict, Optional

from ...graph import Tensor
from .._type._ctype import tffi
from ...enum_defines import Layout
from ...enum_defines import DevType
from .base import Operator, OperatorAttrs, register_op, BaseLayout


class iqCatAttrs(OperatorAttrs):
    def checkparams(self) -> None:
        assert "scale_o" in self.attrs
        assert "dim" in self.attrs

    def serialize(self) -> bytes:
        attrs = tffi.new("iqCatAttrs *")
        attrs.axis = self.attrs["dim"]
        return bytes(tffi.buffer(attrs))


@register_op
class iqCat(Operator, BaseLayout):
    def __init__(self, attrs={}):
        self.attrs = iqCatAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        num_input = len(inputs)

        for i in range(len(inputs)):
            scale_x = self.attrs.get("scale_x_{}".format(i), 1.0)
            temp = math.log(scale_x, 2)
            assert abs(temp - int(temp)) < 0.000001
            assert inputs[i].scale == int(
                temp
            ), "scale in tensor must be same with scale in attribute"

        axis = int(self.attrs["dim"])
        shape = list(inputs[0].shape)
        ndims = len(shape)
        assert axis < ndims and axis >= -ndims
        if axis < 0:
            axis += ndims

        for num in range(1, num_input):
            assert len(inputs[0].shape) == len(inputs[num].shape)
            for i in range(ndims):
                if i != axis:
                    assert inputs[0].shape[i] == inputs[num].shape[i]
                    shape[i] = inputs[0].shape[i]
                else:
                    shape[i] += inputs[num].shape[i]

        scale_o = self.attrs.get("scale_o", 1.0)
        temp = math.log(scale_o, 2)
        assert abs(temp - int(temp)) < 0.000001
        Y = Tensor.clone(inputs[0], shape=tuple(shape), scale=int(temp))

        inputdata = [x.data for x in inputs]
        if all([x.has_data() for x in inputs]):
            inputdata = [x.data for x in inputs]
            Y.data = np.concatenate(inputdata, axis=axis)

        self.outputs = [Y]

    def sub_layout_convert(self):
        inputs = self.inputs
        if inputs[0].layout == Layout.NHWC:
            axis = self.attrs["dim"]
            if axis == 1:
                axis = 3
            elif axis == 2:
                axis = 1
            elif axis == 3:
                axis = 2
            self.attrs["dim"] = axis


__all__ = ["iqCat"]



from ._ctype import tffi
from ...enum_defines import ALIGN16


class tTenDimPair(object):
    def __init__(self, tensor_id, dim_id):
        self.obj = tffi.new("tTenDimPair *")
        self.obj.tensor_id_ = tensor_id
        self.obj.dim_id_ = dim_id

    def to_bytes(self):
        return bytes(tffi.buffer(self.obj))


class tDyAxisInfo(object):
    def __init__(self, tensor_id, dim_id, scalar_id):
        self.obj = tffi.new("tDyAxisInfo *")
        self.obj.tensor_id_ = tensor_id
        self.obj.dy_dim_id_ = dim_id
        self.obj.scalar_input_id_ = scalar_id

    def to_bytes(self):
        return bytes(tffi.buffer(self.obj))


class tShapeInferHdr(object):
    def __init__(self, dy_axis_list, id_pair_list, scalar_graph):
        assert len(id_pair_list) == len(scalar_graph.outputs)
        self.hdr = tffi.new("tShapeInferHdr *")
        pack_buffer = bytes(tffi.buffer(self.hdr))
        offset = ALIGN16(len(pack_buffer))
        pack_buffer += b"\0" * (offset - len(pack_buffer))
        # serialize dynamic axis
        self.hdr.dy_axis_offset_ = offset
        for dy_axis in dy_axis_list:
            pack_buffer += dy_axis.to_bytes()
        offset = ALIGN16(len(pack_buffer))
        pack_buffer += b"\0" * (offset - len(pack_buffer))
        self.hdr.num_dy_axis_ = len(dy_axis_list)

        # serialize scalar graph
        self.hdr.graph_offset_ = offset
        graph_buffer = scalar_graph.serialize()
        pack_buffer += graph_buffer
        offset = ALIGN16(len(pack_buffer))
        pack_buffer += b"\0" * (offset - len(pack_buffer))
        self.hdr.graph_size_ = len(graph_buffer)

        # serialize id pairs
        self.hdr.id_pair_offset_ = offset
        for id_pair in id_pair_list:
            pack_buffer += id_pair.to_bytes()
        offset = ALIGN16(len(pack_buffer))
        pack_buffer += b"\0" * (offset - len(pack_buffer))
        self.hdr.num_id_pair_ = len(id_pair_list)

        hdr_buff = bytes(tffi.buffer(self.hdr))
        self.bytes = hdr_buff + pack_buffer[len(hdr_buff) :]

    def to_bytes(self):
        return self.bytes


__all__ = ["tTenDimPair", "tDyAxisInfo", "tShapeInferHdr"]



import math
import numpy as np

from .utils import QuantType
from ...graph import Tensor
from .._type._ctype import tffi
from ...enum_defines import DevType, Layout, ALIGN2, ALIGN8, ALIGN16
from .utils import attr2tuple, calc_conv1d_output_shape
from .base import Operator, OperatorAttrs, CPUConvLayout, register_op


class Conv1dIntAttrs(OperatorAttrs):
    def checkparams(self) -> None:
        assert "data_bits" in self.attrs
        assert "o_bits" in self.attrs
        assert "parameter_bits" in self.attrs
        assert "kernel_shape" in self.attrs
        assert "pads" in self.attrs
        assert "strides" in self.attrs
        assert "dilations" in self.attrs
        assert "group" in self.attrs
        assert "platform_quant" in self.attrs
        assert "scale_x" in self.attrs
        assert "scale_w" in self.attrs
        assert "scale_o" in self.attrs

        kernels = self.attrs.get("kernel_shape")
        kernels = attr2tuple(kernels, (1,))
        assert len(kernels) == 1, "kernel of Conv1dInt must be one dim"
        assert kernels[0] in {1, 2, 3, 4, 5}, "kernel_w for Conv1dInt exceed limit"

        pads = self.attrs.get("pads")
        pads = attr2tuple(pads, (0, 0))
        assert len(pads) == 2, "pads of conv1dInt must be two dims"
        assert pads[0] in {0, 1, 2, 3, 4}, "pad_left for Conv1dInt exceed limit"
        assert pads[1] in {0, 1, 2, 3, 4}, "pad_right for Conv1dInt exceed limit"

        strides = self.attrs.get("strides")
        strides = attr2tuple(strides, (1,))
        assert len(strides) == 1, "stride_w for Conv1dInt must be one dim"
        assert strides[0] in {1, 2, 4}, "stride_h for Conv1dInt exceed limit"

        assert (
            kernels[0] >= strides[0]
        ), "weight and stride size of Conv1dInt do not match"
        assert (
            pads[0] <= kernels[0] and pads[1] <= kernels[0]
        ), "pad_h and weight_h size of Conv1dInt do not match"

        assert self.attrs.get("platform_quant") == "luna_quant"

    def serialize(self) -> bytes:
        attrs = tffi.new("Conv1dIntAttrs *")

        attrs.kernel = self.attrs["kernel_shape"][0]
        attrs.pad = self.attrs["pads"]
        attrs.stride = self.attrs["strides"][0]
        attrs.group = self.attrs["group"]
        attrs.quant_type = QuantType.from_str(self.attrs["platform_quant"]).value
        attrs.act_type = int(self.attrs.get("act_type", 0))

        return bytes(tffi.buffer(attrs))


@register_op
class Conv1dInt(Operator, CPUConvLayout):
    def __init__(self, attrs={}):
        self.attrs = Conv1dIntAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) == 2 or len(inputs) == 3
        X = inputs[0]
        W = inputs[1]
        assert len(X.shape) == 3
        assert len(W.shape) == 3

        scale_x = self.attrs.get("scale_x")
        temp = math.log(scale_x, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert self.inputs[0].scale == int(temp)

        scale_y = self.attrs.get("scale_w")
        temp = math.log(scale_y, 2)
        assert abs(temp - int(temp)) < 0.000001
        self.inputs[1].scale = int(temp)

        scale_o = self.attrs.get("scale_o")
        temp = math.log(scale_o, 2)
        assert abs(temp - int(temp)) < 0.000001

        kernels = self.attrs.get("kernel_shape")
        strides = self.attrs.get("strides")
        pads = self.attrs.get("pads")

        group = self.attrs.get("group", 1)

        x_c = X.shape[1]
        x_h = X.shape[2]

        assert (
            kernels[0] == W.shape[2]
        ), "kernel_size:{} must same with input of weight:{}".format(
            (kernels[0]), (W.shape[2])
        )
        assert (
            x_h >= W.shape[2]
        ), "input and weight size of convolution do not match".format(
            (x_h), (W.shape[2])
        )
        assert (
            x_c == W.shape[1] * group
        ), "channel in from input:{}, channel in from weight:{}, group:{}".format(
            x_c, W.shape[1], group
        )

        shape = calc_conv1d_output_shape(
            input_shape=X.shape,
            weight_shape=W.shape,
            kernels=kernels,
            strides=strides,
            dilations=(1, 1),
            pads=pads,
            groups=self.attrs["group"],
        )

        assert X.dtype == np.int8, "input data type of convolution must be int8"
        assert W.dtype == np.int8, "weight data type of convolution must be int8"
        if len(inputs) == 3:
            assert (
                inputs[2].dtype == np.int32
            ), "bias data type of convolution must be int32"

        Y = X.clone(shape=tuple(shape), scale=int(temp))
        self.outputs = [Y]

    def get_workspace(self, dev_type: DevType):
        workspace_size = 0
        input_data = self.inputs[0]
        weight = self.inputs[1]
        kernel_num = weight.shape[0]
        kernel_c = weight.shape[1]

        group = self.attrs["group"]
        group_align = ALIGN16(group)
        strides = self.attrs["strides"]
        stride_h = strides[0]
        stride_w = 1
        kernels = self.attrs["kernel_shape"]
        kernel_h = kernels[0]
        kernel_w = 1
        pads = self.attrs["pads"]
        pad_l = pads[0]
        pad_r = pads[1]
        pad_t = 0
        pad_b = 0
        h = input_data.shape[2]
        if len(input_data.shape) == 4:
            w = input_data.shape[3]
        else:
            w = 1
        data_size = (
            ((w + 8 * stride_w - 1) // (8 * stride_w))
            * (8 * stride_w)
            * ((kernel_c + 7) // 8)
            * 8
            * h
        )
        out_size = self.outputs[0].nbytes

        if 1 == kernel_c and (
            group == kernel_num
            or (group_align == kernel_num and weight.layout == Layout.NHWC8)
        ):  # depthwise conv
            num_input_align = kernel_c
            num_output_align = ALIGN16(kernel_num)
            kernel_size = num_output_align * kernel_h * kernel_w
            assert kernel_size * weight.dtype.itemsize <= 32768
            workspace_size = max(out_size, 65536)
        elif group != 1:  # group conv
            assert kernel_num % group == 0
            num_input_align = ALIGN8(kernel_c)
            num_output_align = ((kernel_num // group + 1) // 2) * 2
            kernel_size = num_input_align * num_output_align * kernel_h * kernel_w
            workspace_size = max(out_size, 65536)
        elif data_size > 64 * 1024 and h * w >= 64 * 1024:  # conv split h
            k_h = kernel_h
            s_h = stride_h
            split_num = 1
            tmp_in_h = h
            ou_h = (h + pad_t + pad_b - kernel_h) // stride_h + 1
            input_limit_without_h = (
                ((w + 8 * stride_w - 1) // (8 * stride_w))
                * (8 * stride_w)
                * ((kernel_c + 7) // 8)
                * 8
            )
            while tmp_in_h * input_limit_without_h > 65536 or (ou_h % split_num) != 0:
                split_num = split_num + 1
                tmp_in_h = ((ou_h * s_h) / split_num) + k_h - s_h
            tmp_ou_h = ou_h / split_num
            out_size = math.ceil(out_size / split_num)
            workspace_size = max(out_size, 65536)
        else:
            return []

        max_workspace = Tensor.from_shape([workspace_size], np.int8, dev_type)
        max_workspace.dev_type = dev_type

        return [max_workspace]

    def pack_params(self, dev_type: DevType):
        input_data = self.inputs[0]

        kernels = self.attrs.get("kernel_shape")
        kernel_h = kernels[0]
        kernel_w = 1

        strides = self.attrs.get("strides")
        stride_h = strides[0]
        stride_w = 1

        group = self.attrs.get("group")

        if len(self.inputs) >= 3:
            bias = self.inputs[2]
            assert bias.dtype == np.int32

        weight = self.inputs[1]
        if weight.layout == Layout.NHWC8:
            self.inputs[1].update(
                shape=weight.shape, data=weight.data, layout=Layout.NHWC8
            )
        else:
            kernel_num = weight.shape[0]
            kernel_c = weight.shape[1]
            assert kernel_c * group == input_data.shape[1]
            assert kernel_h == weight.shape[2]
            if len(weight.shape) == 4:
                assert kernel_w == weight.shape[3]

            h = input_data.shape[2]
            if len(input_data.shape) == 4:
                w = input_data.shape[3]
            else:
                w = 1

            weight.data = weight.data.reshape(kernel_num, kernel_c, kernel_h, kernel_w)

            data_size = (
                ((w + 8 * stride_w - 1) // (8 * stride_w))
                * (8 * stride_w)
                * ((kernel_c + 7) // 8)
                * 8
                * h
            )
            assert ~(
                data_size * input_data.dtype.itemsize <= 65536
                and ALIGN8(kernel_c) < 512
            ), "input size of Conv1dInt exceed limit"

            if 1 == kernel_c and group == kernel_num:  # depthwise conv
                num_input_align = kernel_c
                num_output_align = ALIGN16(kernel_num)
                kernel_size = num_output_align * kernel_h * kernel_w
                group = num_output_align
                assert kernel_size * weight.dtype.itemsize <= 32768

                new_weight_data = np.zeros(
                    (num_output_align, num_input_align, kernel_h, kernel_w),
                    weight.dtype,
                )
                shape = new_weight_data.shape
                for p in range(kernel_num):
                    for q in range(kernel_c):
                        new_weight_data[p, q, :, :] = weight.data[p, q, :, :]

                temp = new_weight_data.reshape(
                    -1, 16, num_input_align, kernel_h, kernel_w
                )
                new_weight_data = temp.transpose(0, 2, 3, 4, 1)
                self.inputs[1].update(
                    shape=shape, data=new_weight_data, layout=Layout.NHWC8
                )

            elif group != 1:  # group conv
                assert kernel_num % group == 0
                num_input_align = ALIGN8(kernel_c)
                num_output_align = ((kernel_num // group + 1) // 2) * 2
                kernel_size = num_input_align * num_output_align * kernel_h * kernel_w
                assert kernel_size * weight.dtype.itemsize <= 32768

                new_weight_data = np.zeros(
                    (group, num_output_align, num_input_align, kernel_h * kernel_w),
                    weight.dtype,
                )
                shape = (group * num_output_align, kernel_h, kernel_w, num_input_align)
                weight.data = weight.data.reshape(
                    group, -1, kernel_c, kernel_h * kernel_w
                )

                for g in range(group):
                    for p in range(kernel_num // group):
                        for q in range(kernel_c):
                            new_weight_data[g, p, q, :] = weight.data[g, p, q, :]

                new_weight_data = new_weight_data.transpose(0, 1, 3, 2)
                temp = new_weight_data.reshape(
                    group, -1, 2, kernel_h * kernel_w, num_input_align
                )
                temp1 = temp.transpose(0, 1, 3, 2, 4)
                shape1 = temp1.shape
                temp2 = temp1.reshape(
                    shape1[0], shape1[1] * shape1[2], shape1[3], -1, 8
                )
                new_weight_data = temp2.transpose(0, 1, 3, 2, 4)
                self.inputs[1].update(
                    shape=shape, data=new_weight_data, layout=Layout.NHWC8
                )

            else:  # common conv
                num_input_align = ALIGN8(kernel_c)
                num_output_align = ALIGN2(kernel_num)
                kernel_size = num_input_align * num_output_align * kernel_h * kernel_w
                if (
                    kernel_size * weight.dtype.itemsize > 32768
                ) and data_size * input_data.dtype.itemsize <= 65536:
                    temp_size = num_input_align * kernel_h * kernel_w
                    max_size = 32768 // temp_size
                    if max_size % 2:
                        max_size -= 1
                    split_num = (kernel_num + max_size - 1) // max_size

                    out_size = (kernel_num + split_num - 1) // split_num
                    if out_size % 2:
                        out_size += 1
                    num_output_align = out_size * split_num

                new_weight_data = np.zeros(
                    (num_output_align, num_input_align, kernel_h, kernel_w),
                    weight.dtype,
                )

                for p in range(kernel_num):
                    for q in range(kernel_c):
                        new_weight_data[p, q, :, :] = weight.data[p, q, :, :]
                new_weight_data = new_weight_data.transpose(0, 2, 3, 1)
                shape = new_weight_data.shape
                temp = new_weight_data.reshape(
                    -1, 2, kernel_h, kernel_w, num_input_align
                )
                temp1 = temp.transpose(0, 2, 3, 1, 4)
                shape1 = temp1.shape
                temp2 = temp1.reshape(
                    shape1[0], shape1[1] * shape1[2], shape1[3], -1, 8
                )
                new_weight_data = temp2.transpose(0, 1, 3, 2, 4)
                assert input_data.layout in {Layout.NCHW, Layout.NCWH}
                if input_data.layout == Layout.NCHW:
                    self.inputs[1].update(
                        shape=shape,
                        data=new_weight_data,
                        layout=Layout.NHWC8,
                    )
                elif input_data.layout == Layout.NCWH:
                    self.inputs[1].update(
                        shape=shape, data=new_weight_data, layout=Layout.NWHC8
                    )


__all__ = ["Conv1dInt"]



from .utils.utils import *
from .._type._ctype import tffi
from ...enum_defines import Layout
from ...enum_defines import DevType
from .base import Operator, OperatorAttrs, register_op


class SliceAttrs(OperatorAttrs):
    def serialize(self) -> bytes:
        attrs = tffi.new("SliceAttrs *")

        attrs.axis = self.attrs["axis"]
        attrs.dims = self.attrs["dims"]
        attrs.split = self.attrs["split"]

        return bytes(tffi.buffer(attrs))


@register_op
class Slice(Operator):
    def __init__(self, attrs={}):
        self.attrs = OperatorAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) >= 3
        X = inputs[0]
        starts = inputs[1].data[0]
        ends = inputs[2].data[0]
        if len(inputs) >= 4:
            axes = inputs[3].data[0]
        else:
            axes = 0
        if len(inputs) == 5:
            steps = inputs[4].data[0]
        else:
            steps = 1
        shape = list(X.shape)

        assert axes < len(shape)
        axes = axes + len(shape) if axes < 0 else axes

        if starts < 0:
            starts += shape[axes]

        if ends > 65536:
            ends = shape[axes]
        elif ends < 0:

            ends += shape[axes]

        if steps < 0:
            shape[axes] = (starts - ends + 1 + steps) // (-1 * steps)
        else:
            shape[axes] = (ends - starts + steps - 1) // steps

        Y = X.clone(shape=tuple(shape))
        self.outputs = [Y]

    def sub_layout_convert(self):
        inputs = self.inputs
        if inputs[0].layout == Layout.NHWC:
            axes = inputs[3].data[0]
            if axes == 1:
                axes = 3
            elif axes == 2:
                axes = 1
            elif axes == 3:
                axes = 2
            inputs[3].data[0] = axes


__all__ = ["Slice"]



import math
import numpy as np
from typing import Any, Dict, Optional

from ...graph import Tensor
from ...enum_defines import DevType
from .base import UnaryOperator, OperatorAttrs, register_op


@register_op
class Dequant(UnaryOperator):
    def checkparams(self) -> None:
        assert "scale_o" in self.attrs

    def infer_tensor(self):
        assert len(self.inputs) in {1, 2}
        X = self.inputs[0]

        scale_x = self.attrs.get("scale_o")
        temp = math.log(scale_x, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert X.scale == int(temp)

        Y = X.clone(dtype=np.dtype("f4"), scale=1.0)
        self.outputs = [Y]
        if all([x.has_data() for x in self.inputs]):
            self.forward()


__all__ = ["Dequant"]



import math
import sympy
from enum import Enum


class XEnum(Enum):
    @classmethod
    def from_str(cls, name, default=None):
        name_ = name.name if isinstance(name, Enum) else str(name)
        if name_.upper() in cls.__dict__:
            return cls.__dict__[name_.upper()]
        if default is not None:
            return default
        raise LookupError(f"Unknown enum name: {name} for {cls.__name__}.")

    @classmethod
    def from_int(cls, value):
        value_ = value.value if isinstance(value, Enum) else int(value)
        return cls(value_)


class QuantType(XEnum):
    QSHIFT_QUANT = 0
    QMAX_QUANT = 1
    TF_QUANT = 2
    NORMAL_QUANT = 3
    LUNA_QUANT = 4


class CeilMode(XEnum):
    NO = 0
    CEIL = 1


class AutoPad(XEnum):
    NOTSET = 0
    SAME_UPPER = 1
    SAME_LOWER = 2
    VALID = 3


class PadMode(XEnum):
    CONSTANT = 0
    REFLECT = 1
    EDGE = 2


def attr2tuple(value, default, dim=None):
    # get dim
    if dim == None:
        if isinstance(default, int):
            dim = 1
        else:
            dim = len(default)

    if value is None:
        if isinstance(default, int):
            return tuple(int(default) for i in range(dim))
        return tuple(int(default[i]) for i in range(dim))
    # str2int
    if isinstance(value, str):
        value = eval(value)

    if isinstance(value, int):
        return tuple(int(value) for i in range(dim))
    elif isinstance(value, (list, tuple)):
        if len(value) == 0:
            return tuple(int(default) for i in range(dim))
        elif len(value) == 1:
            return tuple(int(value[0]) for i in range(dim))
        elif len(value) == 2:
            if dim == 2:
                return tuple(int(value[i]) for i in range(dim))
            elif dim == 4:
                return tuple(int(value[i // 2]) for i in range(dim))
            else:
                raise ValueError(f"Invalid attr {value}")
        elif len(value) == 4:
            return tuple(int(value[i]) for i in range(dim))
        elif len(value) == 3:
            return tuple(int(value[i]) for i in range(dim))
        elif len(value) == 6:
            return tuple(int(value[i]) for i in range(dim))
        else:
            raise ValueError(f"Invalid attr {value}")
    else:
        raise ValueError(f"Invalid attr {value}")


__all__ = ["XEnum", "QuantType", "CeilMode", "AutoPad", "PadMode", "attr2tuple"]



from functools import reduce

from .base import Operator, OperatorAttrs, register_op


def _prod(L):
    return reduce(lambda x, y: x * y, L)


@register_op
class Reshape(Operator):
    def __init__(self, attrs={}):
        self.attrs = OperatorAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs

        X = self.inputs[0]
        tShape = list(X.shape)
        yshape = list(inputs[1].data)

        i_1 = None
        for i, s in enumerate(yshape):
            if s == 0:  # copy shape 0
                yshape[i] = tShape[i]
            elif s == -1:  # record shape -1 to i_1
                assert i_1 is None, "Only one -1 accept in reshape"
                yshape[i] = 1
                i_1 = i

        xsize = _prod(tShape)
        ysize = _prod(yshape)
        if i_1 is not None:
            yshape[i_1] = int(xsize / ysize)

        Y = X.clone(shape=tuple(yshape))
        if X.has_data():
            Y.data = X.data.reshape(yshape)
        self.outputs = [Y]

    def is_inplace(self):
        return True


__all__ = ["Reshape"]



import math
import numpy as np
from typing import List

from ...graph import Tensor
from ...enum_defines import DevType
from .base import Operator, register_op


@register_op
class BatchNorm2dInt(Operator):
    def infer_tensor(self):
        assert "scale_x" in self.attrs.attrs
        scale_x = self.attrs.get("scale_x")
        temp = math.log(scale_x, 2)
        assert (
            abs(temp - int(temp)) < 0.000001
        ), "scale of inputs in BatchNorm2dInt must be 2^Q"
        assert self.inputs[0].scale == int(temp)

        assert "scale_w" in self.attrs.attrs
        scale_w = self.attrs.get("scale_w")
        temp = math.log(scale_w, 2)
        assert (
            abs(temp - int(temp)) < 0.000001
        ), "scale of weight in BatchNorm2dInt must be 2^Q"
        self.inputs[1].scale = int(temp)

        self.inputs[2].scale = self.inputs[0].scale + self.inputs[1].scale

        assert "scale_o" in self.attrs.attrs
        scale_o = self.attrs.get("scale_o")
        temp = math.log(scale_o, 2)
        assert (
            abs(temp - int(temp)) < 0.000001
        ), "scale of output in BatchNorm2dInt must be 2^Q"

        Y = self.inputs[0].clone(scale=int(temp))

        self.outputs = [Y]
        return

    def get_workspace(self, dev_type: DevType) -> List[Tensor]:
        X = self.inputs[0]
        workspace_bytes = X.shape[2] * X.shape[3] * 4
        max_workspace = Tensor.from_shape([workspace_bytes], np.int8, dev_type)
        return [max_workspace]


__all__ = ["BatchNorm2dInt"]



from .._type._ctype import tffi
from .base import Operator, OperatorAttrs, register_op


class ResizeAttrs(OperatorAttrs):
    def checkparams(self) -> None:
        self.attrs["cubic_coeff_a"] = self.attrs.get("cubic_coeff_a", -0.75)
        self.attrs["coordinate_transformation_mode"] = self.attrs.get(
            "coordinate_transformation_mode", "half_pixel"
        )
        self.attrs["mode"] = self.attrs.get("mode", "nearest")
        self.attrs["nearest_mode"] = self.attrs.get(
            "nearest_mode", "round_prefer_floor"
        )

        coord_trans_mode = self.attrs["coordinate_transformation_mode"]
        if coord_trans_mode == "half_pixel":
            coord_trans_mode = 0
        elif coord_trans_mode == "pytorch_half_pixel":
            coord_trans_mode = 1
        elif coord_trans_mode == "align_corners":
            coord_trans_mode = 2
        elif coord_trans_mode == "asymmetric":
            coord_trans_mode = 3
        else:  # tf_crop_and_resize
            coord_trans_mode = 4
        self.attrs["pack_coord_trans_mode"] = coord_trans_mode

        self.attrs["cubic_coeff_a"] = self.attrs["cubic_coeff_a"]

        mode = self.attrs["mode"]
        if mode == "nearest":
            mode = 0
        elif mode == "linear":
            mode = 1
        else:  # cubic
            mode = 2
        self.attrs["pack_mode"] = mode

        nearest_mode = self.attrs["nearest_mode"]
        if nearest_mode == "round_prefer_floor":
            nearest_mode = 0
        elif nearest_mode == "round_prefer_ceil":
            nearest_mode = 1
        elif nearest_mode == "floor":
            nearest_mode = 2
        else:  # ceil
            nearest_mode = 3
        self.attrs["pack_nearest_mode"] = nearest_mode

    def serialize(self) -> bytes:
        attrs = tffi.new("ResizeAttrs *")

        attrs.coord_trans_mode = self.attrs["coordinate_transformation_mode"]
        attrs.cubic_coeff_a = self.attrs["cubic_coeff_a"]
        attrs.mode = self.attrs["mode"]
        attrs.nearest_mode = self.attrs["nearest_mode"]

        return bytes(tffi.buffer(attrs))


@register_op
class Resize(Operator):
    def __init__(self, attrs={}):
        self.attrs = ResizeAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        X = inputs[0]
        Y = None
        if inputs[2].nbytes != 0:
            shape = X.shape * inputs[2].data
            yshape = []
            for i, s in enumerate(shape):
                yshape.append(int(shape[i]))

            Y = X.clone(shape=tuple(yshape))
        elif inputs[3].nbytes != 0:
            shape = inputs[3].data
            Y = X.clone(shape=tuple(shape))
        self.outputs = [Y]


__all__ = ["Resize"]



from .activations import *
from .BatchNorm2dInt import *
from .BmmInt import *
from .Clip import *
from .Conv1dInt import *
from .Conv2dInt import *
from .ConvTranspose2dInt import *
from .Dequant import *
from .Flatten import *
from .Gather import *
from .GRUInt import *
from .iqAdd import *
from .iqCat import *
from .iqDiv import *
from .iqMul import *
from .iqSigmoid import *
from .iqSub import *
from .iqVar import *
from .LayerNormInt import *
from .LinearInt import *
from .LogSoftmaxInt import *
from .LstmInt import *
from .Pool import *
from .Quant import *
from .Requant import *
from .Reshape import *
from .Resize import *
from .ShuffleChannel import *
from .Slice import *
from .SoftmaxInt import *
from .Split import *
from .Squeeze import *
from .Transpose import *
from .Unsqueeze import *



import math
import numpy as np
from typing import Any, Dict, Optional

from ...graph import Tensor

from .._type._ctype import tffi
from .utils import QuantType
from ...enum_defines import DevType, MemType, ALIGN4, ALIGN16, ALIGN32
from .base import Operator, OperatorAttrs, register_op


class LinearIntAttrs(OperatorAttrs):
    def checkparams(self) -> None:
        assert "scale_x" in self.attrs
        assert "scale_w" in self.attrs
        assert "scale_o" in self.attrs
        assert "data_bits" in self.attrs
        assert "o_bits" in self.attrs
        assert "parameter_bits" in self.attrs
        assert "platform_quant" in self.attrs

        self.attrs["transA"] = self.attrs.get("transA", 0)
        self.attrs["transB"] = self.attrs.get("transB", 1)
        self.attrs["alpha"] = self.attrs.get("alpha", 1.0)
        self.attrs["beta"] = self.attrs.get("beta", 0.0)

    def serialize(self) -> bytes:
        attrs = tffi.new("LinearIntAttrs *")

        attrs.transA = self.attrs["transA"]
        attrs.transB = self.attrs["transB"]
        quant_type = QuantType.from_str(self.attrs.get("platform_quant", "qmax_quant"))

        attrs.quant_type = quant_type.value

        return bytes(tffi.buffer(attrs))


@register_op
class LinearInt(Operator):
    def __init__(self, attrs={}):
        self.attrs = LinearIntAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) in {2, 3}
        X = inputs[0]
        W = inputs[1]
        x_shape = list(X.shape)
        w_shape = list(W.shape)
        assert X.dtype == W.dtype

        x_h = x_shape[-2]
        x_w = x_shape[-1]

        scale_x = self.attrs.get("scale_x")
        temp = math.log(scale_x, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert self.inputs[0].scale == int(temp)

        scale_w = self.attrs.get("scale_w")
        temp = math.log(scale_w, 2)
        assert abs(temp - int(temp)) < 0.000001
        self.inputs[1].scale = temp

        scale_o = self.attrs.get("scale_o")
        temp = math.log(scale_o, 2)
        assert abs(temp - int(temp)) < 0.000001
        self.outputs[0].scale = temp

        if not self.attrs["transA"] and self.attrs["transB"]:
            assert x_w == w_shape[-1], "x_w:{} and w_shape[-1]:{}".format(
                x_w, w_shape[-1]
            )
        elif self.attrs["transA"] and self.attrs["transB"]:
            assert x_h == w_shape[-1], "x_h:{} and w_shape[-1]:{}".format(
                x_h, w_shape[-1]
            )
        elif not self.attrs["transA"] and not self.attrs["transB"]:
            assert x_w == w_shape[0], "x_w:{} and w_shape[0]:{}".format(x_w, w_shape[0])
        else:
            assert x_h == w_shape[0]

        M = ALIGN16(x_h)
        N = ALIGN32(x_w)
        if X.dtype == np.int8:
            assert M * N < 65536, "left matmul of linearint must less 64KB"
        elif X.dtype == np.int16:
            assert M * N < 32768, "left matmul of linearint must less 64KB"
        elif X.dtype == np.int32:
            assert M * N < 16384, "left matmul of linearint must less 64KB"
        else:
            raise (f"[LinearInt] Not supported type of X.dtype:{X.dtype}.")

        # 1D
        if len(X.shape) == 1:
            shape = [w_shape[0]]
        elif len(X.shape) == 2:
            shape = [x_shape[0], w_shape[0]]
        else:
            shape = [x_shape[0], x_shape[1], w_shape[0]]

        Y = X.clone(shape=shape, dtype=X.dtype, scale=int(temp))

        self.outputs = [Y]

    def get_workspace(self, dev_type: DevType):
        if len(self.inputs) > 2:
            workspace_bytes = self.outputs[0].nbytes * self.inputs[2].dtype.itemsize
            max_workspace = Tensor.from_shape([workspace_bytes], np.int8, dev_type)
            return [max_workspace]
        elif self.inputs[0].mem_type != self.outputs[0].mem_type:
            workspace_bytes = self.outputs[0].nbytes * self.outputs[0].dtype.itemsize
            max_workspace = Tensor.from_shape([workspace_bytes], np.int8, dev_type)
            return [max_workspace]
        else:
            return []

    def pack_params(self, dev_type: DevType):
        transA = self.attrs["transA"]
        transB = self.attrs["transB"]
        if transB:
            weight = self.inputs[1]
            data = weight.data.transpose(1, 0)
            self.inputs[1].update(data=data, shape=data.shape)
            # self.attrs["transB"] = 0


__all__ = ["LinearInt"]



import math
import numpy as np

from .utils import QuantType
from .._type._ctype import tffi
from ...enum_defines import DevType
from typing import Any, Dict, List, Optional
from .base import Operator, OperatorAttrs, register_op


class RequantAttrs(OperatorAttrs):
    def checkparams(self) -> None:
        assert "scale_x" in self.attrs
        assert "scale_o" in self.attrs
        assert "data_bits" in self.attrs
        assert "o_bist" in self.attrs

    def serialize(self) -> bytes:
        attrs = tffi.new("RequantAttrs *")

        attrs.scale_x = self.attrs["scale_x"]
        attrs.scale_o = self.attrs["scale_o"]
        attrs.o_bits = self.attrs["o_bits"]
        attrs.data_bits = self.attrs["data_bits"]
        attrs.quant_type = self.attrs["quant_type"].value

        return bytes(tffi.buffer(attrs))


@register_op
class Requant(Operator):
    def __init__(self, attrs={}):
        self.attrs = OperatorAttrs(attrs)

    def infer_tensor(self):
        X = self.inputs[0]

        scale_x = self.attrs.get("scale_x")
        temp1 = math.log(scale_x, 2)
        assert temp1 == int(temp1)
        assert self.inputs[0].scale == int(temp1)

        scale_o = self.attrs.get("scale_o")
        temp1 = math.log(scale_o, 2)
        assert temp1 == int(temp1)

        if 32 == self.attrs["o_bits"]:
            data_type = np.dtype("i4")
        elif 16 == self.attrs["o_bits"]:
            data_type = np.dtype("i2")
        elif 8 == self.attrs["o_bits"]:
            data_type = np.dtype("i1")
        else:
            raise ZeroDivisionError(f"o_bits is not equal data_bits.")

        Y = X.clone(dtype=data_type, scale=int(temp1))
        self.outputs = [Y]


__all__ = ["Requant"]



from .base import Operator, OperatorAttrs, register_op


@register_op
class Flatten(Operator):
    def __init__(self, attrs={}):
        self.attrs = OperatorAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) == 1
        X = inputs[0]

        shape = list(X.shape)
        size = 1
        for i in range(1, len(shape)):
            size *= shape[i]
        Y_shape = (shape[0], size)
        Y = X.clone(shape=tuple(Y_shape))
        self.outputs = [Y]


__all__ = ["Flatten"]



from ...graph import Tensor
from .._type._ctype import tffi
from .base import Operator, OperatorAttrs, register_op


class ShuffleChannelAttrs(OperatorAttrs):
    def checkparams(self) -> None:
        self.attrs["axis"] = self.attrs.get("axis", 1)
        assert "groups" in self.attrs

    def serialize(self) -> bytes:
        attrs = tffi.new("ShuffleChannelAttrs *")

        attrs.num_group = self.attrs["groups"]
        attrs.axis = self.attrs["axis"]

        return bytes(tffi.buffer(attrs))


@register_op
class ShuffleChannel(Operator):
    def __init__(self, attrs={}):
        self.attrs = ShuffleChannelAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) == 1
        X = inputs[0]
        assert len(X.shape) == 4
        Y = Tensor.clone(self.inputs[0])
        self.outputs = [Y]


__all__ = ["ShuffleChannel"]



import numpy as np

from .._type import tffi
from .base import Operator, OperatorAttrs, BaseLayout, register_op


class ClipAttrs(OperatorAttrs):
    def serialize(self) -> bytes:
        attrs = tffi.new("ClipAttrs")

        assert "min" in self.attrs.attrs
        assert "max" in self.attrs.attrs
        attrs.min = self.attrs["min"]
        attrs.max = self.attrs["max"]

        return bytes(tffi.buffer(attrs))


@register_op
class Clip(Operator, BaseLayout):
    def __init__(self, attrs={}):
        self.attrs = ClipAttrs(attrs)

    def infer_tensor(self):
        assert len(self.inputs) == 1 or len(self.inputs) == 3
        X = self.inputs[0]
        Y = X.clone()
        if all([x.has_data() for x in self.inputs]):
            min = self.inputs[1].data
            max = self.inputs[2].data

            Y.data = np.clip(self.inputs[0].data, min, max)
        self.outputs = [Y]


__all__ = ["Clip"]



from ..graph import Graph
from .simple_fusion import MethodManager
from ..save_model import save_to_onnx_model


def op_fusion(
    graph: Graph,
    strategy: str,
    is_dump: bool = True,
    dump_file_path: str = "./model.ignore/graph_op_fusion.onnx",
) -> Graph:

    """
    Args:
      graph: Graph
        待融合graph
      ignore_methods: List[str]
        忽略特定算子融合;
      is_dump: bool
        是否保存融合后的graph，默认保存，即保存；
      dump_file_path: str
        保存路径，默认保存到./graph_sim_fusion.onnx
    Returns:
      g: Graph
        融合后的graph
    """
    ignore_methods = list()
    if strategy != None:
        ignore_methods = [strategy]

    graph_after_fusion = MethodManager.apply(graph, ignore_methods)

    if is_dump:
        save_to_onnx_model(graph_after_fusion, dump_file_path)
    return graph_after_fusion


__all__ = ["op_fusion"]



import math

from ...enum_defines import DevType
from .base import iqBinaryOperator, register_op


@register_op
class iqMul(iqBinaryOperator):
    def infer_tensor(self):
        X = self.inputs[0]
        scale_x = self.attrs.get("scale_x")
        temp = math.log(scale_x, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert self.inputs[0].scale == temp

        scale_y = self.attrs.get("scale_y")
        temp = math.log(scale_y, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert self.inputs[1].scale == temp

        scale_o = self.attrs.get("scale_o")
        temp = math.log(scale_o, 2)
        assert abs(temp - int(temp)) < 0.000001

        Y = X.clone(scale=int(temp))


__all__ = ["iqMul"]



import math
import numpy as np

from ...graph import Tensor
from .._type._ctype import tffi
from ...enum_defines import DevType
from .base import Operator, OperatorAttrs, register_op


class SoftmaxIntAttrs(OperatorAttrs):
    def serialize(self) -> bytes:
        attrs = tffi.new("SoftmaxIntAttrs *")
        attrs.axis = self.attrs["dim"]
        return bytes(tffi.buffer(attrs))


@register_op
class SoftmaxInt(Operator):
    def __init__(self, attrs={}):
        self.attrs = SoftmaxIntAttrs(attrs)

    def infer_tensor(self):
        inputs = self.inputs
        assert len(inputs) == 1
        X = inputs[0]

        scale_x = self.attrs.get("scale_x")
        temp = math.log(scale_x, 2)
        assert abs(temp - int(temp)) < 0.000001
        assert X.scale == temp

        scale_o = self.attrs.get("scale_o")
        temp = math.log(scale_o, 2)
        assert abs(temp - int(temp)) < 0.000001

        Y = X.clone(scale=int(temp))
        self.outputs = [Y]

    def get_workspace(self, dev_type: DevType):
        axis = self.attrs.get("dim", -1)
        workspace_bytes = self.inputs[0].shape[axis] * 2
        max_workspace = Tensor.from_shape([workspace_bytes], np.int32, dev_type)
        return [max_workspace]


__all__ = ["SoftmaxInt"]



from typing import List
from collections import defaultdict

from ...graph import Graph


class MethodManager(object):
    """method管理类
    管理所有method，并封装了method apply接口
    类成员变量：
      methods: Dict[str, func]
        用于存储op融合方式；
    """

    methods = defaultdict()

    @classmethod
    def apply(cls, g: Graph, ignore_methods: List[str]) -> Graph:

        """遍历所有的methods进行算子融合
        Args:
          g: Graph
            待融合graph
          specify_methods: List[str]
            只指定特定算子融合, 其他方式将忽略；
          ignore_methods: List[str]
            忽略特定算子融合；
        Returns:
          g: Graph
            融合后的graph
        """
        support_methods = list(cls.methods.keys())
        methods = set(support_methods)
        if len(ignore_methods) > 0:
            methods.difference_update(set(ignore_methods))

        for method_name in support_methods:
            if method_name in methods:
                g = cls.methods[method_name](g)

        return g

    @classmethod
    def remove(cls, name):
        """删除指定method"""
        del cls.methods[name]

    @classmethod
    def clear(cls):
        """清空所有的method"""
        cls.methods.clear()


def register_method(name):
    """用于注册method的装饰器"""

    def wrapper(cls):
        MethodManager.methods[name] = cls
        return cls

    return wrapper


__all__ = ["MethodManager", "register_method"]



import os
from cffi import FFI

tffi = FFI()

_ctype_file1 = os.path.dirname(__file__) + "/../../executor/core/operator_attrs.h"
_ctype_file2 = (
    os.path.dirname(__file__) + "/../../executor/include/thinker/thinker_type.h"
)

with open(_ctype_file1) as f:
    _ctype_cdef = f.readlines()
    _ctype_cdef = [
        x
        for x in _ctype_cdef
        if not x.startswith("#ifndef")
        and not x.startswith("#endif")
        and not x.startswith("#include")
    ]

tffi.cdef("\n".join(_ctype_cdef))

with open(_ctype_file2) as f:
    _ctype_cdef = f.readlines()
    _ctype_cdef = [
        x
        for x in _ctype_cdef
        if not x.startswith("#ifndef")
        and not x.startswith("#endif")
        and not x.startswith("#include")
    ]

tffi.cdef("\n".join(_ctype_cdef))

__all__ = ["tffi"]



import numpy as np
from ._ctype import tffi
from ...enum_defines import ALIGN16


class tOperator(object):
    def __init__(
        self,
        op_attrs,
        op_type,
        dev_type,
        num_input,
        num_output,
        tensor_ids=[],
        scalar_ids=[],
    ):

        self.op_type = op_type
        self.tensor_ids = tensor_ids
        self.scalar_ids = scalar_ids
        self.attrs_buff = bytes(op_attrs)

        self.obj = tffi.new("tOperator *")
        self.obj.op_id_ = 0

        self.obj.num_input_ = num_input
        self.obj.num_output_ = num_output
        self.obj.num_temp_ = len(tensor_ids) - num_input - num_output
        # self.obj.num_scalar_ = len(scalar_ids)
        self.obj.num_scalar_ = dev_type.value

    def pack(self):
        pack_buff = bytes(tffi.buffer(self.obj))
        offset = len(pack_buff)

        self.obj.attr_offset_ = offset
        pack_buff += self.attrs_buff
        offset = ALIGN16(len(pack_buff))
        pack_buff += b"\0" * (offset - len(pack_buff))

        self.obj.tensor_offset_ = offset
        pack_buff += np.array(self.tensor_ids, dtype=np.int32).tobytes()
        offset = ALIGN16(len(pack_buff))
        pack_buff += b"\0" * (offset - len(pack_buff))

        self.obj.scalar_offset_ = offset
        self.obj.total_size_ = offset

        hdr_buff = bytes(tffi.buffer(self.obj))
        self.bytes = hdr_buff + pack_buff[len(hdr_buff) :]

    def to_bytes(self):
        self.pack()
        return self.bytes


class tOperatorList(object):
    def __init__(self, op_list):
        self._list = op_list
        self.obj = tffi.new("tOperatorList *")

        op_types = {}
        op_type_list = []
        op_buff = b""
        for x in op_list:
            if x.op_type not in op_types:
                op_types[x.op_type] = len(op_type_list)
                op_type_list.append(x.op_type)
            x.obj.op_id_ = op_types[x.op_type]  # update op_id
            op_buff += x.to_bytes()
        if len(op_type_list) != 0:
            max_len = max([len(x) for x in op_type_list])
            max_len = ALIGN16(max_len + 1)
        else:
            max_len = 0

        self.obj.op_count_ = len(op_list)
        self.obj.type_count_ = len(op_types)

        self.obj.type_length_ = max_len
        self.obj.header_size_ = tffi.sizeof("tOperatorList")
        self.obj.type_offset_ = ALIGN16(self.obj.header_size_)
        self.obj.op_offset_ = self.obj.type_offset_ + max_len * len(op_types)
        self.obj.op_size_ = len(op_buff)

        self.bytes = bytes(tffi.buffer(self.obj))
        for name in op_type_list:
            self.bytes += np.array([ord(x) for x in name], dtype=np.uint8).tobytes()
            self.bytes += b"\0" * (max_len - len(name))

        self.bytes += op_buff

    def to_bytes(self):
        return self.bytes


_all__ = ["tOperator", "tOperatorList"]



from ....enum_defines import Layout


def ProdShape(Shape):
    result = 1
    for x in Shape:
        result = result * x
    return result


def gemm_workspace_desc(
    matA, is_a_trans, is_a_pack, matB, is_b_trans, is_b_pack, dtype_size
):
    M = 0
    N = 0
    K = 0
    if False == is_a_trans and False == is_b_trans:
        M = matA[0]
        K = matA[1]
        N = matB[1]
    elif False == is_a_trans and True == is_b_trans:
        M = matA[0]
        K = matA[1]
        N = matB[0]
    elif True == is_a_trans and False == is_b_trans:
        M = matA[1]
        K = matA[0]
        N = matB[1]
    elif True == is_a_trans and True == is_b_trans:
        M = matA[1]
        K = matA[0]
        N = matB[0]

    workspace_bytes = 0
    if False == is_a_pack and M > 1:
        workspace_bytes += dtype_size * M * K
    if False == is_b_pack and N > 1:
        workspace_bytes += dtype_size * K * N
    return workspace_bytes


def get_deconv_workspace_desc(calc_kernel, X, W, attrs, outputs):
    group = attrs["group"]
    layout = attrs["layout"]
    input_shape = X.shape
    weight_shape = W.shape
    dtype_size = X.dtype.itemsize

    M = 0
    N = 0
    K = 0
    workspace_size = 0
    matA = []
    matB = []
    is_a_trans = 0
    is_b_trans = 0
    if layout == Layout.NCHW:
        M = ProdShape(weight_shape[1:])
        N = ProdShape(input_shape[2:])
        K = weight_shape[0] / group
        is_a_trans = 1
        matA = [K, M]
        is_b_trans = 0
        matB = [K, N]
    else:
        M = ProdShape(input_shape[1:3])
        N = ProdShape(weight_shape[1:])
        K = weight_shape[0] / group
        is_a_trans = 0
        matA = [M, K]
        is_b_trans = 0
        matB = [K, N]
        if group > 1:
            workspace_size += dtype_size * M * K

    workspace_size += dtype_size * M * N
    workspace_size += gemm_workspace_desc(
        matA, is_a_trans, False, matB, is_b_trans, False, dtype_size
    )
    return int(workspace_size)


def get_gemm_workspace_desc(calc_kernel, inputs, attrs, outputs):
    matA = inputs[0].shape
    matB = inputs[1].shape
    dtype_size = inputs[0].dtype.itemsize

    is_a_trans = attrs["transA"]
    is_b_trans = attrs["transB"]

    workspace_size = gemm_workspace_desc(
        matA, is_a_trans, False, matB, is_b_trans, False, dtype_size
    )
    return workspace_size


__all__ = [
    "gemm_workspace_desc",
    "get_deconv_workspace_desc",
    "get_gemm_workspace_desc",
]



from .Graph import *
from .Tensor import *
from .indexed import *
from .Tensor import *
from .GraphNode import *
from .GraphEntry import *
from .GraphEntryExt import *

__all__ = [
    "Tensor",
    "GraphEntry",
    "ConstantEntry",
    "ScalarEntry",
    "InputEntry",
    "OutputEntry",
    "EmptyEntry",
    "GraphNode",
    "Graph",
]



# Copyright (C) 2022 listenai Co.Ltd
# All rights reserved. 
# Created by leifang on 2022.09.31

from enum import Enum
from typing import List


def ALIGN2(x):
    return (x + 1) >> 1 << 1


def ALIGN4(x):
    return (x + 3) >> 2 << 2


def ALIGN8(x):
    return (x + 7) >> 3 << 3


def ALIGN16(x):
    return (x + 15) >> 4 << 4


def ALIGN32(x):
    return (x + 31) >> 5 << 5


class DevType(Enum):
    HIFI = 0
    LUNA = 1

    @staticmethod
    def from_str(dev: str) -> "DevType":
        if dev.upper() == "HIFI":
            return DevType.HIFI
        elif dev.upper() == "LUNA":
            return DevType.LUNA
        assert 0, "{} is not found in dev list".format(dev)


class Platform(Enum):
    VENUS = 0

    @staticmethod
    def from_str(platform: str) -> "Platform":
        if platform.upper() == "VENUS":
            return Platform.VENUS
        assert 0, "{} is not found in Platform list".format(platform)

    def get_cpu_list(platform: str) -> List[DevType]:
        if platform.upper() == "VENUS":
            return [DevType.HIFI, DevType.LUNA]

    def get_support_op(platform: str) -> List[str]:
        if platform.upper() == "VENUS":
            return [
                "Conv1dInt",
                "Conv2dInt",
                "ConvTranspose2dInt",
                "Relu",
                "Prelu",
                "Clip",
                "MaxPool",
                "AvgPool2dInt",
                "LinearInt",
                "GRUInt",
                "LSTMInt",
                "BmmInt",
                "iqSigmoid",
                "iqTanh",
                "iqCat",
                "Transpose",
                "Reshape",
                "Squeeze",
                "Unsqueeze",
                "Flatten",
                "Slice",
                "iqSum",
                "iqAdd",
                "iqMul",
                "iqDiv",
                "Resize",
                "Gather",
                "Quant",
                "Dequant",
                "SoftmaxInt",
                "ShuffleChannel",
                "LogSoftmaxInt",
                "Split",
                "BatchNorm2dInt",
                "ReduceMean",
                "Requant",
                "LayerNormInt",
                "iqVar",
            ]


class MemType(Enum):
    FLASH = 0
    PSRAM = 1
    SHARE_MEM = 2

    @staticmethod
    def from_str(dev: str) -> "MemType":
        if dev.upper() == "FLASH":
            return MemType.FLASH
        elif dev.upper() == "PSRAM":
            return MemType.PSRAM
        elif dev.upper() == "SHARE-MEM":
            return MemType.SHARE_MEM
        assert 0, "{} is not found in dev".format(dev)


class Layout(Enum):
    NCHW = 0
    NHWC = 1
    NCWH = 2
    NHWC8 = 3
    NWHC8 = 4

    @staticmethod
    def from_str(layout_str: str) -> "Layout":
        if layout_str == "NCHW":
            return Layout.NCHW
        if layout_str == "NHWC":
            return Layout.NHWC
        if layout_str == "NCWH":
            return Layout.NCWH
        if layout_str == "NHWC8":
            return Layout.NHWC8
        if layout_str == "NWHC8":
            return Layout.NWHC8


class TensorType(Enum):
    Normal = 0
    Input = 2
    Output = 3
    Emptry = 4


class LayoutPerfData:
    def __init__(self, kernel: str = None, performance: int = 10):
        self.kernel = kernel
        self.performance = performance
        self.inputs_layout = []
        self.outputs_layout = []


__all__ = [
    "ALIGN2",
    "ALIGN4",
    "ALIGN8",
    "ALIGN16",
    "ALIGN32",
    "DevType",
    "Platform",
    "MemType",
    "Layout",
    "TensorType",
    "LayoutPerfData",
]



# Copyright (C) 2022 listenai Co.Ltd
# All rights reserved. 
# Created by leifang on 2022.09.31

import os
import sympy
import onnx
import numpy as np
from typing import AnyStr, Optional

from onnx.numpy_helper import from_array
from onnx.helper import make_tensor_value_info
from onnx.helper import make_node, make_graph, make_model
from onnx.mapping import NP_TYPE_TO_TENSOR_TYPE

from .graph import Graph


def _normalize_attrs(node):
    from enum import Enum

    new_attrs = {}
    for name, attr in node.attrs.items():
        if isinstance(attr, Enum):
            new_attrs[name] = attr.name
        elif name == "to" and node.op_type == "Cast":
            new_attrs[name] = _convert_dtype(node.attrs["to"])
        else:
            if name != "graph":
                new_attrs[name] = attr

    return new_attrs


def _convert_dtype(dtype):
    """Convert thinker tensor dtype to onnx tensor dtype"""
    np_dtype = np.dtype(dtype)
    onnx_dtype = NP_TYPE_TO_TENSOR_TYPE.get(np_dtype)
    if onnx_dtype is None:
        raise TypeError(f"Unsupported dtype: {dtype}!")
    return onnx_dtype


def _convert_shape(shape):
    """Convert thinker tensor shape to onnx tensor shape"""
    ret = list()
    for i in shape:
        if isinstance(i, sympy.Symbol):
            ret.append(i.name)
        elif isinstance(i, int) or isinstance(i, str):
            ret.append(i)
        elif isinstance(i, np.int64):
            ret.append(int(i))
        else:
            raise TypeError(f"Unsupported dim type:{type(i)}!")
    return ret


def _convet_to_onnx_model(
    thinkerGraph: Graph, graph_name: Optional[str] = "", **model_attrs
) -> onnx.ModelProto:
    """Convert thinker graph to onnx model
    Due to all Constant Node has been conveted to GraphEntry.THINKER_TENSOR_CONSTANT,
    Constant Node couldn't been rebuilt in the target onnx model.
    """
    onnx_inputs = list()
    onnx_outputs = list()
    onnx_initializer = list()
    onnx_nodes = list()

    # sort
    thinkerGraph.update()

    # graph input
    for inp in thinkerGraph.inputs:
        entry = inp.tensor
        dtype = _convert_dtype(entry.dtype)
        shape = _convert_shape(entry.shape)
        tensor_vi = make_tensor_value_info(inp.name, dtype, shape)
        onnx_inputs.append(tensor_vi)

    # graph output
    for out in thinkerGraph.outputs:
        onnx_outputs.append(onnx.ValueInfoProto(name=out.name))

    # graph node
    for node in thinkerGraph.nodes.values():
        new_attrs = _normalize_attrs(node)
        for key in new_attrs.keys():
            if isinstance(new_attrs[key], np.dtype):
                new_attrs[key] = _convert_dtype(new_attrs[key])
        onnx_node = make_node(
            op_type=node.op_type,
            inputs=[i.name for i in node.inputs],
            outputs=[i.name for i in node.outputs],
            name=node.name,
            **new_attrs,
        )
        onnx_nodes.append(onnx_node)

    # graph initializer
    for entry in thinkerGraph.entries.values():
        if entry.is_constant():
            tensor = entry.tensor
            if tensor.data is None:
                raise ValueError(f"tensor.data is None!")
            onnx_constant = from_array(tensor.data, name=entry.name)
            onnx_initializer.append(onnx_constant)

    # make graph
    onnx_graph = make_graph(
        onnx_nodes, graph_name, onnx_inputs, onnx_outputs, onnx_initializer
    )
    save_model_attrs = {}
    save_model_attrs.update(model_attrs)
    onnx_model = make_model(
        onnx_graph, producer_name="thinker", producer_version="v1", **save_model_attrs
    )
    return onnx_model


def save_to_onnx_model(
    thinkerGraph: Graph, path: AnyStr, name: Optional[AnyStr] = "", **model_attrs
):
    """save thinker graph to onnx model"""
    graph = thinkerGraph.clone(is_update=False)
    onnx_model = _convet_to_onnx_model(graph, name, **model_attrs)

    # create folder
    str_lists = path.split("/")
    temp = ""
    for i in range(len(str_lists) - 1):
        temp += str_lists[i] + "/"
    if os.path.exists(temp) == False and temp != "":
        os.makedirs(temp)
    try:
        onnx.save(onnx_model, path)
    except ValueError:
        pass


__all__ = ["save_to_onnx_model"]



# Copyright (C) 2022 listenai Co.Ltd
# All rights reserved. 
# Created by leifang on 2022.09.31

from typing import List, Dict

from .graph import Graph
from .resource_packer import (
    pack_param,
    pack_memory,
    pack_tensor,
    pack_operator,
    pack_io,
)
from .resource_packer import (
    tModel,
    tMemoryList,
    tTensorList,
    tOperatorList,
    tIOInfo,
    tParameterList,
    tDebugList,
    tDMAList,
)


def serialize(graph: Graph, memory_plan: Dict[int, List[int]]):
    print("pack param begin")
    param_list, shared_memory_list = pack_param(memory_plan)
    print("pack param success")
    print("pack tensor begin")
    runtime_memory_list = pack_memory(memory_plan)
    tensor_list, tensor_name_list = pack_tensor(memory_plan)
    print("pack tensor success")
    print("pack operator begin")
    operator_list, dma_list = pack_operator(graph, memory_plan)
    print("pack operator success")
    print("pack io begin")
    input_list, output_list = pack_io(graph)
    print("pack io success")
    packed_model = tModel(
        tMemoryList(shared_memory_list, runtime_memory_list),
        tTensorList(tensor_list),
        tOperatorList(operator_list),
        tIOInfo(input_list, output_list),
        tParameterList(param_list),
        tDebugList(tensor_name_list),
        tDMAList(dma_list),
    )
    return packed_model


__all__ = ["serialize"]
